(load "getopt.lisp")

(require :sb-posix)
(require :sb-bsd-sockets)

(defconstant color-d (format nil "~c[0m"    #\escape))
(defconstant color-r (format nil "~c[0;31m" #\escape))
(defconstant color-g (format nil "~c[0;32m" #\escape))

(defconstant data-dir-def "/usr/local/etc/greasynet")
(defconstant interval-def "100")

(defconstant rpcc-temp    "/tmp/gnconfig.XXXXXX")
(defconstant rpcc-file    "rpc")
(defconstant rpcs-file    "rpc")

(defconstant sock-rcvbuf 65536)
(defconstant sock-sndbuf 65536)

(defconstant cmd-args
  '((help     sym     )
    (stop             )
    (if       str     )
    (subnet   str     )
    (ip       id  str )
    (port     itg     )
    (id       id  id  )
    (locaddr  id  adr )
    (extaddr  id  adr )
    (addr     id  adr )
    (mute     id  bol )
    (brd      id  bol )
    (add      id      )
    (del      id      )
    (addkey   id  str )
    (delkey   id  str )
    (genkey           )
    (rekey    id      )
    (prod     id      )
    (net      id      )
    (stat     id      )
    (keys     id      )
    (fkeys    id      )))

(defun p (fmt &rest arg)
  (apply #'format t fmt arg))

(defun cls()
  (p "~c[1J~c[H" #\escape #\escape))

(defun path-cat (&rest arg)
  (if (cdr arg)
    (concatenate 'string (car arg) "/" (apply #'path-cat (cdr arg)))
    (car arg)))

(defun must-exist (path)
  (if (probe-file path)
    path
    (error "file '~a' doesn't exist" path)))

(defun rm-f (path)
  (if (probe-file path)
    (delete-file path)))

(defun sock-unix-dgram ()
  (let ((sock (make-instance 'sb-bsd-sockets:local-socket
                             :type :datagram)))
    (setf (sb-bsd-sockets:sockopt-receive-buffer  sock) sock-rcvbuf)
    (setf (sb-bsd-sockets:sockopt-send-buffer     sock) sock-sndbuf)
    sock))

(defun sock-read (sock)
  (read-from-string
    (sb-bsd-sockets:socket-receive sock nil sock-rcvbuf)))

(defun sock-write (sock data)
  (let ((str (write-to-string data)))
    (sb-bsd-sockets:socket-send sock str (length str))))

(defun sym (x &optional (name "argument"))
  (let ((sym (read-from-string x)))
    (if (symbolp sym)
      sym
      (error "~a must be symbol" name))))

(defun str (x) x)

(defun itg (x)
  (handler-case (parse-integer x)
    (error (e) (error "argument must be integer"))))

(defun bol (x)
  (cond
    ((string= x "on")   t)
    ((string= x "off")  nil)
    (t                  (error "argument must be on or off"))))

(defun id (x)
  (if (zerop (length x))
    ""
    (let ((sym (read-from-string x)))
      (if (eql sym 'self) sym x))))

(defun adr (x)
  (let ((cp (position #\: x)))
    (if cp
     `((ip    . ,(subseq x 0 cp))
       (port  . ,(handler-case (parse-integer (subseq x (1+ cp)))
                   (error (e) (error "address port must be integer"))))
       (stamp . 0))
      (error "address must be in format a.b.c.d:port"))))

(defun parse-arg (cmd arg)
  (let ((types (assoc cmd cmd-args)))
    (unless types (error "unknown rpc command '~(~a~)'" cmd))
    (setf types (cdr types))
    (loop for a in arg
          for tcdr = types then (cdr tcdr)
          for type = (or (car tcdr) 'str)
      collect (ecase type
                ('sym (sym a))
                ('str (str a))
                ('itg (itg a))
                ('bol (bol a))
                ('id  (id  a))
                ('adr (adr a))))))

(defconstant default-reqs '((net) (keys) (stat)))

(defun sep-args (arg)
  (if arg
    (let ((idx (position "-" arg :test #'string=)))
      (if idx
        (cons (subseq arg 0 idx)
              (sep-args (subseq arg (1+ idx))))
        (list arg)))))

(defun parse-reqs (arg)
  (if arg
    (loop for a in (delete nil (sep-args arg))
      collect (let* ((cmd (sym (car a) "command")))
                (cons cmd (parse-arg cmd (cdr a)))))
    default-reqs))

(defun alu (sym alist)
  (cdr (assoc sym alist)))

(defun id-self (id)
  (and (symbolp id) (eql id 'self)))

(defun fmt-id (id)
  (cond
    ((id-self id) "<self>")
    ((stringp id) id)
    (t            (error (e) "invalid id"))))

(defun fmt-ipport (ip port)
  (format nil "~a:~a" ip port))

(defun fmt-addr (addr)
  (if addr
    (fmt-ipport (alu 'ip    addr)
                (alu 'port  addr))
    "-"))

(defun fmt-flag (flag chr)
  (if flag chr #\_))

(defun fmt-net-flags (n)
  (format nil "~c~c"
    (fmt-flag (alu 'mute n) #\m)
    (fmt-flag (alu 'brd  n) #\b)))

(defconstant stamp-off (encode-universal-time 0 0 0 1 1 1970 0))

(defun stamp-to-uni (stamp)
  (+ stamp stamp-off))

(defun uni-to-stamp (uni)
  (- uni stamp-off))

(defun fmt-uni (uni)
  (multiple-value-bind (ts tm th dd dm dy yd ds tz)
                       (decode-universal-time uni)
    (format nil "~2,'0d-~2,'0d-~2,'0d ~2,'0d:~2,'0d:~2,'0d"
            dy dm dd th tm ts)))

(defun fmt-stamp (stamp)
  (if (zerop stamp)
    "-"
    (fmt-uni (stamp-to-uni stamp))))

(defconstant sec-min  (*  60 1))
(defconstant sec-hour (*  60 sec-min))
(defconstant sec-day  (*  24 sec-hour))
(defconstant sec-year (* 365 sec-day))

(defun fmt-stamp-diff (idiff)
  (let* ((sign  (if (minusp idiff) "-" ""))
         (diff  (abs idiff))
         (unit  nil))
    (cond
      ((< diff sec-min)   (setf                             unit "s"))
      ((< diff sec-hour)  (setf diff (floor diff sec-min)   unit "m"))
      ((< diff sec-day)   (setf diff (floor diff sec-hour)  unit "h"))
      ((< diff sec-year)  (setf diff (floor diff sec-day)   unit "d"))
      (t                  (setf diff (floor diff sec-year)  unit "y")))
    (format nil "~a~d~a"
            sign diff unit)))

(defun fmt-ago (stamp ref)
  (if (zerop stamp)
    (format nil "~5@a" "-")
    (let* ((ago   (- ref stamp))
           (long  (>= ago 30)))
      (format nil "~a~a~a"
              (if long "" color-g)
              (format nil "~5@a" (fmt-stamp-diff ago))
              (if long "" color-d)))))

(defun clamp (n lo hi)
  (max (min n hi) lo))

(defconstant size-pfix  '("B" "KB" "MB" "GB" "TB" "PB" "EB"))
(defconstant size-dec   100)

(defun fmt-size (size)
  (let* ((x (clamp (floor (if (zerop size)
                            0
                            (log size 2))
                          10)
                   1
                   (1- (length size-pfix))))
         (y (expt 2 (* 10 x)))
         (usize (floor (* size size-dec) y))
         (isize (floor usize size-dec))
         (msize (mod   usize size-dec)))
    (format nil "~d.~v,'0d ~a"
            isize
            (log size-dec 10)
            msize
            (nth x size-pfix))))

(defun handle-response (res)
  (let* ((cmd (car    res))
         (arg (cdr    res))
         (a0  (first  arg))
         (a1  (second arg))
         (a2  (third  arg)))

    (case cmd
      (err      (error "~a" a0))
      (ok)

      (help-all (p "~12a ~a~%" "command" "description")
                (dolist (c a0)
                  (p "~(~12a~) ~a~%" (car c) (cdr c))))

      (help-cmd (p "~a~%"         (alu 'desc  a0))
                (p "usage: ~a~%"  (alu 'usage a0)))

      (if       (p "~a~%" a0))
      (subnet   (p "~a~%" a0))
      (ip       (p "~a~%" a1))
      (port     (p "~d~%" a0))
      (id       (p "~a~%" a1))

      ((locaddr extaddr)
        (p "~a~%" (fmt-addr a1)))

      ((mute brd)
        (p "~a~%" (if a1 "on" "off")))

      (genkey   (p "~a~%" a0))

      (net
        (p "~10@<id~> ~15@<virtual ip~> ~21@<effective address~> ~
            ~4@<flag~> ~20<timestamp~> ~5<seen~>~%")
        (dolist (n a1)
          (let* ((id        (alu 'id n))
                 (effaddr   (alu 'effaddr n))
                 (effstamp  (if effaddr (alu 'stamp effaddr) 0)))
            (p "~10a ~15a ~21a ~4a ~20@a ~a~%"
               (fmt-id    id)
               (alu 'ip   n)
               (fmt-addr  effaddr)
               (if (id-self id) "" (fmt-net-flags n))
               (fmt-stamp effstamp)
               (fmt-ago   effstamp a0)))))

      ((stat-total stat-peer)
        (p "~28@<metric (~a)~> ~
            ~14<rx packets~> ~14<tx packets~> ~
            ~10<rx bytes~> ~10<tx bytes~>~%"
            (if (eql cmd 'stat-total)
              "total"
              (fmt-id (alu 'id a0))))
        (labels
         ((fsym (fmt &rest arg)
            (read-from-string (apply #'format nil fmt arg)))

          (stat (s)
            (p "~28a" (cdr s))
            (dolist (dir '("rx" "tx"))
              (p " ~14d" (alu (fsym "~a~an" (car s) dir) a0)))
            (dolist (dir '("rx" "tx"))
              (p " ~10@a" (fmt-size (alu (fsym "~a~ab" (car s) dir) a0))))
            (p "~%")))

          (mapcar #'stat '(("g" . "greasypak udp")
                           ("n" . "encapsulated ipv4")
                           ("r" . "rejected udp(rx)/ipv4(tx)")
                           ("d" . "discarded ipv4(rx)/udp(tx)")))))

      ((keys fkeys)  
        (p "~10@<id~> ~v@<key~a~> ~8<rxseq~> ~8<txseq~> ~
            ~v<timestamp~> ~5<seen~>~%"
           (if (eql cmd 'keys) 8 128)
           (if (eql cmd 'keys) " crc" "")
           (if (eql cmd 'keys) 36 19))
        (dolist (k a1)
          (p "~10a ~va ~8,'0x ~8,'0x ~v@a ~a~%"
            (fmt-id (alu 'id k))
            (if (eql cmd 'keys) 8 128)
            (alu 'key   k)
            (alu 'rxseq k)
            (alu 'txseq k)
            (if (eql cmd 'keys) 36 19)
            (fmt-stamp  (alu 'stamp k))
            (fmt-ago    (alu 'stamp k) a0))))

      (otherwise
        (error "don't know how to format rpc response '~(~a~)'" cmd)))))

(defun send-reqs (reqs sock)
  (do ((r reqs (cdr r)))
      ((null r))
    (sock-write sock (car r))
    (handle-response (sock-read sock))
    (if (cdr r) (p "~%"))))

(defun monitor (reqs sock interval)
  ;in case eror dont cls fuk up stderr
  (send-reqs reqs sock)
  (loop
    (cls)
    (send-reqs reqs sock)
    (sleep interval)))

(defun usag ()
  (p "usage: gnconfig [-d dir] [-m [-i interval_ms]] ~
      [cmd_line] [- cmd_line]...~%~
      cmd_line: rpc_cmd [arg...]~%~
      default cmd_lines: net - keys - stat~%~
      run 'gnconfig help' to list all rpc commands~%~
      options:~%  ~
      -h        show help~%  ~
      -d dir    data dir (default = '~a')~%  ~
      -m        monitor command~%  ~
      -i int    monitor refresh interval in ms (default = ~a)~%"
     data-dir-def
     interval-def))

(defun main ()
  (handler-case
    (getopt argv ((help     "h")
                  (data-dir "d" :def data-dir-def)
                  (monitor  "m")
                  (interval "i" :def interval-def))
      (setf interval (/ (parse-integer interval) 1000))
      (if help
        (usag)
        (let* ((reqs     (parse-reqs argv))
               (rpcc-dir (sb-posix:mkdtemp rpcc-temp)))
          (unwind-protect
            (let ((rpcc-path  (path-cat rpcc-dir rpcc-file))
                  (rpcs-path  (must-exist (path-cat data-dir rpcs-file)))
                  (sock       (sock-unix-dgram)))
              (sb-bsd-sockets:socket-bind sock rpcc-path)
              (unwind-protect
                (progn
                  (sb-bsd-sockets:socket-connect sock rpcs-path)
                  (if monitor
                    (monitor reqs sock interval)
                    (send-reqs reqs sock)))
                (rm-f rpcc-path)))
            (sb-posix:rmdir rpcc-dir)))))

    (sb-sys:interactive-interrupt ()
      (format *error-output* "~%")
      (sb-ext:exit :code 0))

    (error (e)
      (format *error-output* "error: ~a~%" e)
      (sb-ext:exit :code 1))))

