#include "daemon.h"
#include "internal.h"
#include "pak.h"
#include "rpc/cmd.h"
#include "state/state.h"
#include "peer/peer.h"
#include <def.h>
#include <os.h>
#include <rpc/server.h>
#include <io/fs.h>
#include <net/util.h>
#include <util/time.h>
#include <log/log.h>

#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include <dirent.h>
#include <poll.h>
#include <sys/wait.h>
#include <sys/stat.h>

int             rng_slow_fd = -1;
int             rng_fast_fd = -1;

uid_t           uid         = 0;
gid_t           gid         = 0;

static uint64_t tick_last   = 0;

static bool     running;
static int      exit_code;
static int      exit_sig;

enum{
  IDX_DEV,
  IDX_UDP,
  IDX_RPC,
  POLL_FD_N,
};

static struct pollfd poll_fd[POLL_FD_N];

static void daemon_start(void){
  running = 1;
}

void daemon_stop(int code){
  running   = 0;
  exit_code = code;
}

static void term(int sig){
  exit_sig = sig;
  daemon_stop(EXIT_SUCCESS);
}

static void sig_ignore(void){
  signal(SIGHUP,  SIG_IGN);
  signal(SIGINT,  SIG_IGN);
  signal(SIGPIPE, SIG_IGN);
  signal(SIGTERM, SIG_IGN);
}

static void sig_setup(void){
  signal(SIGINT,  term);
  signal(SIGTERM, term);
}

static bool mkdir_chown(char const *path){
  if(fs_exists(path)) return 1;

  if(!fs_mkdir_p(path,DATA_MODE)){
    return 0;
  }

  if(!fs_chown(path,uid,gid)){
    return 0;
  }

  return 1;
}

static bool mk_dirs(void){
  umask(~ETC_MODE);
  if(!fs_mkdir_p(ETC_PATH,ETC_MODE)) return 0;
  umask(~DATA_MODE);
  if(!mkdir_chown (data_path)) return 0;
  if(!fs_chdir    (data_path)) return 0;
  if(!mkdir_chown (PEER_PATH)) return 0;
  return 1;
}

static bool log_init(void){
  log_close();
  log_add(stderr,LOG_FLAGS_STDERR);
  if(!log_add_file(LOG_PATH,
                   LOG_FLAGS_FILE,
                   DATA_MODE,
                   uid,
                   gid)){
    return 0;
  }
  return 1;
}

static bool perm_fix_f(char const *path){
  if(!fs_chown(path,uid,gid))   return 0;
  if(!fs_chmod(path,DATA_MODE)) return 0;
  return 1;
}

static bool perm_fix(void){
  DIR            *d;
  struct dirent  *de;

  if(!perm_fix_f(CFG_PATH)) return 0;

  if((d = opendir(PEER_PATH)) == 0){
    log_errno("opendir");
    log_err("failed to open directory for listing %s",PEER_PATH);
    return 0;
  }

  while(de = readdir(d)){
    char path[PATH_BUF];
    if(strcmp(de->d_name,"." ) == 0) continue;
    if(strcmp(de->d_name,"..") == 0) continue;
    snprintf(path,sizeof(path),"%s/%s",PEER_PATH,de->d_name);
    if(!perm_fix_f(path)) return 0;
  }

  closedir(d);
  return 1;

fail:
  closedir(d);
  return 0;
}

static bool drop_priv(void){
  if(gid != 0 && setgid(gid) != 0){
    log_errno("setgid");
    log_err("failed to set group id %d",gid);
    return 0;
  }

  if(uid != 0 && setuid(uid) != 0){
    log_errno("setuid");
    log_err("failed to set user id %d",uid);
    return 0;
  }

  if(uid != 0 && setuid(0) != -1){
    log_err("user privileges could be restored after dropping");
    return 0;
  }

  if(gid != 0 && setgid(0) != -1){
    log_err("group privileges could be restored after dropping");
    return 0;
  }

  if(uid != 0 || gid != 0){
    log_out("dropped privileges (uid=%d, gid=%d)",uid,gid);
  }

  if(getuid() == 0 || getgid() == 0){
    log_out("WARNING: child process running as root user or group. "
            "specifying both -u and -g options recommended");
  }

  return 1;
}

static bool rng_init(void){
  rng_slow_fd = rng_open(RNG_SLOW_PATH);
  if(rng_slow_fd == -1){
    return 0;
  }
  log_out("slow rng %s opened",RNG_SLOW_PATH);

  rng_fast_fd = rng_open(RNG_FAST_PATH);
  if(rng_fast_fd == -1){
    close(rng_slow_fd);
    rng_slow_fd = -1;
    return 0;
  }
  log_out("fast rng %s opened",RNG_FAST_PATH);

  return 1;
}

static void rng_exit(void){
  if(rng_close(rng_slow_fd) == 0){
    log_out("closed slow rng");
    rng_slow_fd = -1;
  }

  if(rng_close(rng_fast_fd) == 0){
    log_out("closed fast rng");
    rng_fast_fd = -1;
  }
}

static bool rpcserv_init(void){
  lisp_try{
    sym_self = sym("SELF");

    for(size_t x = 0;x < rpc_cmd_n;x++){
      rpc_cmd_t      *c = rpc_cmd+x;
      rpc_cmdinfo_t  *i = rpc_cmdinfo+x;

      c->sym  = sym(i->sym_name);
      c->func = i->func;
    }
  }lisp_catch{
    log_lisperr(); 
    return 0;
  }

  if(!rpcs_init(RPCS_PATH,rpc_cmd,rpc_cmd_n)) return 0;
  return 1;
}

static void poll_init_in(size_t idx,int fd){
  poll_fd[idx].fd     = fd;
  poll_fd[idx].events = POLLIN;
}

static void poll_init(void){
  poll_init_in(IDX_DEV,dev_fd);
  poll_init_in(IDX_UDP,udp_fd);
  poll_init_in(IDX_RPC,rpc_fd);
}

static int poll_pls(uint64_t ntime){
  if(ntime-tick_last > TICK_INTERVAL_MAX){
    return 0;
  }
  return poll(poll_fd,POLL_FD_N,POLL_TIMEOUT);
}

static bool poll_readable(size_t idx){
  return poll_fd[idx].revents & POLLIN;
}

static int child(void){
  char ipv4_buf   [IPV4_BUF];
  char subnet_buf [SUBNET_BUF];

  log_out("process forked (pid=%d)",getpid());
  exit_code = EXIT_FAILURE;
  exit_sig  = 0;

  if(!drop_priv     ()) goto ret;
  if(!rng_init      ()) goto ret;
  if(!udp_init      ()) goto rng_exit;
  if(!rpcserv_init  ()) goto udp_exit;
  if(!peer_load_all ()) goto rpcs_exit;

#define IP(V) ipv4_fmt(ipv4_buf,sizeof(ipv4_buf),V)
#define PR(V) subnet_fmt(subnet_buf,sizeof(subnet_buf),V)

  log_out(PROG_FULLNAME " running");
  log_out("system         %s",  OS_NAME);
  log_out("interface      %s",  state.ifa);
  log_out("subnet         %s",  PR(state.subnet));
  log_out("virtual ip     %s",  IP(state.ip));
  log_out("udp port       %d",  state.port);
  log_out("external ip    %s",  IP(state.extaddr.ip));
  log_out("external port  %d",  state.extaddr.port);
  log_out("peer count     %zu", peer_n);

  poll_init();
  sig_setup();
  daemon_start();
  prod_all(PROD_REQANS);

  while(running){
    uint64_t ntime = mono_nsec();

    switch(poll_pls(ntime)){
    case -1:
      if(errno == EINTR){
        running = 0;
        break;
      }
      log_errno("poll");
      log_err("failed to poll file descriptors");
      break;

    case 0:
      tick_last = ntime;
      prod_tick       (ntime); 
      addr_tick       (ntime); 
      rekey_tick      (ntime); 
      flood_tick      (ntime); 
      state_flush     ();
      peer_flush_all  ();
      break;

    default:
      if(poll_readable(IDX_DEV)) dev_recv   ();
      if(poll_readable(IDX_UDP)) udp_recv   ();
      if(poll_readable(IDX_RPC)) rpcs_recv  ();
      break;
    }
  }

  if(exit_code == EXIT_RESTART){
    log_out(PROG_NAME " restarting");
  }else switch(exit_sig){
  case SIGINT:
    log_out(PROG_NAME " shuting down (interrupted)");
    break;

  case SIGTERM:
    log_out(PROG_NAME " shuting down (terminated)");
    break;

  default:
    log_out(PROG_NAME " shuting down");
  }

rpcs_exit:
  rpcs_exit();
udp_exit:
  udp_exit();
rng_exit:
  rng_exit();
data_flush:
  state_flush();
  peer_flush_all();
  peer_init_all();
ret:
  return exit_code;
}

int daemon_run(int argc,char **argv){
  lisp_frame_t  frame;
  pid_t         pid;
  int           code = EXIT_FAILURE;

  if(getuid() != 0){
    log_err("daemon must be executed as root");
    goto ret;
  }

  sig_ignore();

  lisp_init();
  lisp_store(&frame);

  state_init();

  if(!mk_dirs     ()) goto pop_frame;
  if(!log_init    ()) goto pop_frame;
restart:
  if(!state_load  ()) goto pop_frame;
  if(!dev_init    ()) goto log_close;
  if(!if_up       ()) goto dev_exit;
  if(!route_setup ()) goto dev_exit;
  if(!perm_fix    ()) goto dev_exit;

  switch(pid = fork()){
    int status;

  case -1:
    log_errno("fork");
    log_err("failed to fork process");
    goto dev_exit; 

  case 0:
    code = child();
    dev_exit(0);
    log_close();
    lisp_load(&frame);
    return code;

  default:
    dev_exit(1);
    if(waitpid(pid,&status,0) == -1){
      log_errno("waitpid");
      log_err("failed to wait for child process");
      goto if_down;
    }
    code = WEXITSTATUS(status);
    log_out("child process exit with status %d",code);
    goto if_down;
  }

dev_exit:
  dev_exit(0);
if_down:
  if_down();
  if(code == EXIT_RESTART){
    code = EXIT_FAILURE;
    goto restart;
  }
log_close:
  log_close();
pop_frame:
  lisp_load(&frame);
ret:
  return code;
}

