#include "internal.h"
#include "pak.h"
#include "state/state.h"
#include <config.h>
#include <log/log.h>
#include <util/util.h>
#include <util/time.h>

#define FP_SIZE     (1<<FLOOD_BUF_POW2)
#define RB_SIZE     (FP_SIZE<<FLOOD_DEDUP_SHIFT)

#define BYTE_BITS   8

typedef struct fp{
  uint64_t  ntime;
  uint8_t   peer  [CEIL_DIV(PEER_MAX,BYTE_BITS)];
  uint8_t   pak   [FLOOD_BUFSIZE];
  size_t    pak_size;
}fp_t;

typedef struct rb{
  uint64_t  hash;
}rb_t;

fp_t        fp[FP_SIZE] = {0};
rb_t        rb[RB_SIZE] = {0};
uintptr_t   rb_f        =  0;
uintptr_t   rb_b        = -1;

static inline size_t div_8(size_t x){
  return x>>3;
}

static inline size_t mod_8(size_t x){
  return x & 8-1;
}

static bool peer_get(fp_t const *f,peer_t const *p){
  size_t idx = peer_idx(p);
  return f->peer[div_8(idx)]>>mod_8(idx) & 0x01;
}

static void peer_set(fp_t *f,peer_t const *p){
  size_t idx = peer_idx(p);
  f->peer[div_8(idx)] |= 0x01<<mod_8(idx);
}

static void peer_clr(fp_t *f){
  memset(f->peer,0x00,sizeof(f->peer));
}

static inline bool flood_empty(void){
  return rb_b == (uintptr_t)-1;
}

static inline size_t flood_n(void){
  if(flood_empty()) return 0;
  return 1+(rb_f > rb_b ? rb_b+RB_SIZE-rb_f :
                          rb_b        -rb_f);
}

static inline size_t fp_wrap(size_t idx){
  return idx & FP_SIZE-1;
}

static inline fp_t *fp_get(size_t off){
  return fp+fp_wrap(off);
}

static inline size_t rb_wrap(size_t idx){
  return idx & RB_SIZE-1;
}

//1 = if found and added peer, 0 = not found
static bool flood_lookup(uint64_t hash,peer_t const *p){
  for(size_t x = 0;x < RB_SIZE;x++){
    rb_t *r = rb+x;
    if(hash != r->hash) continue;
    peer_set(fp_get(x),p);
    return 1;
  }
  return 0;
}

static bool flood_push(uint64_t       hash,
                       peer_t const  *p,
                       uint64_t       ntime,
                       uint8_t       *pak,
                       size_t         pak_size){

  rb_t *r;
  fp_t *f;

  if(flood_n() >= FP_SIZE){
    return 1;
  }

  //TODO: delet old elt from hash tabl if hash!=0

  rb_b        = rb_wrap(rb_b+1);
  r           = rb+rb_b;
  r->hash     = hash;
  f           = fp_get(rb_b);
  f->ntime    = ntime;
  f->pak_size = pak_size;

  peer_clr(f);
  peer_set(f,p);
  memcpy(f->pak,pak,pak_size);

  //TODO: imsert hash tabl

  return 1;
}

static fp_t *flood_pop(void){
  fp_t *f = fp_get(rb_f);
  rb_f++;
  if(rb_b+1 == rb_f){
    rb_f  =  0;
    rb_b  = -1;
  }else{
    rb_f  = rb_wrap(rb_f);
  }
  return f;
}

static bool tx_flood_maybe(uint8_t const *f,size_t f_size,peer_t *p){
  pkey_t *pk;
  addr_t *addr;

  if(!peer_cansend(p,&pk,&addr,0)){
    return 0;
  }

  return tx_load(f,
                 f_size,
                 PAK_TYPE_FLOOD,
                 p,
                 pk,
                 addr_to_in(*addr));
}

static bool front_expired(uint64_t ntime){
  fp_t *f = fp_get(rb_f);
  return ntime >= f->ntime+FLOOD_DELAY;
}

static void fp_relay(fp_t *f){
  for(size_t x = 0;x < peer_n;x++){
    peer_t *p = peer+x;
    if(peer_get(f,p)) continue;
    tx_flood_maybe(f->pak,f->pak_size,p);
  }
}

bool flood_recv(peer_t *p,uint8_t *f,size_t f_maxsize){
  uint8_t  *b = f+FLOOD_LOAD_OFF;
  uint32_t  src;
  uint32_t  dst;
  size_t    b_size;
  size_t    f_size;
  uint8_t   ttl;
  uint64_t  hash;
  peer_t   *q;

  if(f_maxsize < FLOOD_MINSIZE){
    log_err("flood packet from peer %s invalid size (%u)",
            p->id,f_maxsize);
    return 0;
  }

  src     = get_u32le (f+FLOOD_SRC_OFF);
  dst     = get_u32le (f+FLOOD_DST_OFF);
  b_size  = get_u16le (f+FLOOD_SIZE_OFF);
  ttl     = get_u8le  (f+FLOOD_TTL_OFF);
  f_size  = FLOOD_SIZE(b_size);

  if(!ipv4_insubnet(src,state.subnet)){
    log_err("flood packet from peer %s source not in subnet",p->id);
    return 0;
  }

  if(!ipv4_insubnet(dst,state.subnet)){
    log_err("flood packet from peer %s destination not in subnet",p->id);
    return 0;
  }

  if(b_size > FLOOD_LOAD_SIZE(f_maxsize)){
    log_err("flood packet from peer %s invalid load size (%u)",p->id,b_size);
    return 0;
  }

  if(ttl == 0){
    log_err("flood packet from peer %s zero ttl",p->id);
    return 0;
  }

  if(dst == state.ip){
    rx_blk(b,b_size,ADDR_IN_CONST(0,0),1);
    return 1;
  }

  ttl--;
  if(ttl == 0) return 1;
  set_u8le(f+FLOOD_TTL_OFF,ttl);

  q = peer_lookup_ip(dst);
  if(q &&
     !peer_cold(q,time_sec()) &&
     tx_flood_maybe(f,f_size,q)){
    return 1;
  }

  assert(PAK_SIG_SIZE <= sizeof(uint64_t));
  hash = get_u64le(b+BLK_SIG_OFF);

  if(flood_lookup(hash,p)){
    return 1;
  }

  return flood_push(hash,p,mono_nsec(),f,f_size);
}

bool flood_brd(peer_t *p,uint8_t type,uint8_t const *load,size_t load_size){
  uint8_t     f[FLOOD_BUFSIZE];
  uint8_t    *b = f+FLOOD_LOAD_OFF;
  uint8_t    *r = b+BLK_RED_OFF;
  uint8_t    *l = r+RED_LOAD_OFF;
  size_t      b_size;
  size_t      f_size;
  pkey_t     *p_pk;
  size_t      n = 0;
  char const *err;

  if(!peer_cancrypt(p,&p_pk,&err)){
    log_err("%s for peer %s, cannot encrypt flood packet",err,p->id);
    return 0;
  }

  load_copy(l,load,load_size);
  if(!blk_prepare(b,load_size,type,p_pk,&b_size)){
    log_err("failed to prepare flood packet load");
    return 0;
  }

  if(!FLOOD_LOAD_VALIDSIZE(b_size)){
    log_err("invalid flood packet load size");
    return 0;
  }

  set_u32le (f+FLOOD_SRC_OFF,   state.ip);
  set_u32le (f+FLOOD_DST_OFF,   p->ip);
  set_u16le (f+FLOOD_SIZE_OFF,  b_size);
  set_u8le  (f+FLOOD_TTL_OFF,   FLOOD_TTL);
  set_u8le  (f+FLOOD_PAD_OFF,   0x00);

  f_size = FLOOD_SIZE(b_size);

  for(size_t x = 0;x < peer_n;x++){
    peer_t *q = peer+x;
    if(q == p) continue;
    n += tx_flood_maybe(f,f_size,q);
  }

  if(n > 0) pkey_used(p,p_pk);
  return 1;
}

void flood_tick(uint64_t ntime){
  while(!flood_empty() && front_expired(ntime)){
    fp_relay(flood_pop());
  }
}

void flood_flush(void){
  while(!flood_empty()){
    fp_relay(flood_pop());
  }
}

