#define _DEFAULT_SOURCE 1

#include "daemon.h"
#include "internal.h"
#include "pak.h"
#include "state/state.h"
#include <os.h>
#include <net/util.h>
#include <io/fs.h>
#include <log/log.h>
#include <util/string.h>

#include <stdlib.h>
#include <errno.h>
#include <assert.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <net/if.h>
 
#define IBUF_SIZE 32768

static int ioctl_socket(void){
  int fd = socket(AF_INET,SOCK_DGRAM,0);
  if(fd == -1){
    log_errno("socket");
    log_err("failed to create socket");
  }
  return fd;
}

static bool if_setname(int fd){
  struct ifreq  ifr;
  char const   *ifa;

  ifa = dev_ifname();
  if(strcmp(ifa,state.ifa) == 0) return 1;

  VAR_ZERO(ifr);
  strzcpy(ifr.ifr_name,ifa,sizeof(ifr.ifr_name));

#ifdef OS_LINUX
  strzcpy(ifr.ifr_newname,state.ifa,sizeof(ifr.ifr_newname));
#else
	ifr.ifr_data = (caddr_t)state.ifa;
#endif
  if(ioctl(fd,SIOCSIFNAME,(char *)&ifr) == -1){
    log_errno("ioctl");
    log_err("failed to set name of interface %s",ifa);
    return 0;
  }

  return 1;
}

static bool if_setaddr(int fd){
#ifdef OS_LINUX
  struct ifreq ifr;

  VAR_ZERO(ifr);
  strzcpy(ifr.ifr_name,state.ifa,sizeof(ifr.ifr_name));

  set_sockaddr_ip(&ifr.ifr_addr,state.ip);
  if(ioctl(fd,SIOCSIFADDR,(char *)&ifr) == -1){
    log_errno("ioctl");
    log_err("failed to set interface %s address",state.ifa);
    return 0;
  }

  set_sockaddr_ip(&ifr.ifr_netmask,ipv4_maskip(state.subnet.mask));
  if(ioctl(fd,SIOCSIFNETMASK,(char *)&ifr) == -1){
    log_errno("ioctl");
    log_err("failed to set interface %s netmask",state.ifa);
    return 0;
  }

  ifr.ifr_flags = IFF_UP;
  if(ioctl(fd,SIOCSIFFLAGS,(char *)&ifr) == -1){
    log_errno("ioctl");
    log_err("failed to set interface %s flags",state.ifa);
    return 0;
  }

#else
  struct ifaliasreq ifra;

  VAR_ZERO(ifra);
  strzcpy(ifra.ifra_name,state.ifa,sizeof(ifra.ifra_name));

  set_sockaddr_ip(&ifra.ifra_addr,      state.ip);
  set_sockaddr_ip(&ifra.ifra_broadaddr, state.ip);
  set_sockaddr_ip(&ifra.ifra_mask,      ipv4_maskip(state.subnet.mask));
  if(ioctl(fd,SIOCAIFADDR,(char *)&ifra) == -1){
    log_errno("ioctl");
    log_err("failed to set interface %s addresses",state.ifa);
    return 0;
  }
#endif
  return 1;
}

static bool if_setmtu(int fd){
  struct ifreq ifr;

  VAR_ZERO(ifr);
  strzcpy(ifr.ifr_name,state.ifa,sizeof(ifr.ifr_name));

  ifr.ifr_mtu = IP_PAK_MAXSIZE;
  if(ioctl(fd,SIOCSIFMTU,(char *)&ifr) == -1){
    log_errno("ioctl");
    log_err("failed to set mtu of interface %s",state.ifa);
    return 0;
  }

  return 1;
}

bool if_up(void){
  int fd = ioctl_socket();
  if(!fd) goto fail;

  if(!if_setname(fd)) goto close;
  if(!if_setaddr(fd)) goto close;
  if(!if_setmtu(fd))  goto close;

  close(fd);
  log_out("interface %s created",state.ifa);
  return 1;

close:
  close(fd);
fail:
  return 0;
}

static bool if_destroy(int fd){
#ifdef OS_LINUX
  return 1;
#else
  struct ifreq ifr;

  VAR_ZERO(ifr);
  strzcpy(ifr.ifr_name,state.ifa,sizeof(ifr.ifr_name));

  if(ioctl(fd,SIOCIFDESTROY,(char *)&ifr) == -1){
    log_errno("ioctl");
    log_err("failed to destroy interface %s",state.ifa);
    return 0;
  }
  return 1;
#endif
}

bool if_down(void){
  int fd = ioctl_socket();
  if(!fd) goto fail;

  if(!if_destroy(fd)) goto close;
  log_out("interface %s destroyed",state.ifa);

  close(fd);
  return 1;

close:
  close(fd);
fail:
  return 0;
}

bool if_enumbrd(uint32_t *brd,size_t brd_max,size_t *brd_n){
  int             fd;
  struct ifconf   ifc;
  char            ibuf[IBUF_SIZE];
  char           *ifrpc;
  char           *ifend;
  char           *ifnext;

  fd = ioctl_socket();
  if(!fd) goto fail;

  ifc.ifc_len = sizeof(ibuf);
  ifc.ifc_buf = (void *)ibuf;
  if(ioctl(fd,SIOCGIFCONF,(char *)&ifc) == -1){
    log_errno("ioctl");
    log_err("failed to enumerate interfaces");
    goto close;
  }

  *brd_n  = 0;
  ifend   = ibuf+ifc.ifc_len;

  for(ifrpc = ibuf;ifrpc < ifend;ifrpc = ifnext){
    size_t              n;
    struct ifreq       *ifrp;
    struct ifreq        ifr;
    struct sockaddr_in *sa;

    ifrp = (struct ifreq *)ifrpc;
#ifdef OS_LINUX
    ifnext = ifrpc+sizeof(struct ifreq);
    (void)n;
#else
    n = ifrp->ifr_addr.sa_len+sizeof(ifrp->ifr_name);
    if(n < sizeof(struct ifreq)){
      ifnext = ifrpc+sizeof(struct ifreq);
    }else{
      ifnext = ifrpc+n;
    }
    if(ifrp->ifr_addr.sa_family != AF_INET)  continue;
#endif

    VAR_ZERO(ifr);
    strzcpy(ifr.ifr_name,ifrp->ifr_name,sizeof(ifr.ifr_name));

    if(ioctl(fd,SIOCGIFFLAGS,(char *)&ifr) == -1){
      if(errno == ENXIO) continue;
      log_errno("ioctl");
      log_err("failed to get interface flags");
      continue;
    }

    if((ifr.ifr_flags & IFF_UP) == 0) continue;

    if(ioctl(fd,SIOCGIFBRDADDR,(char *)&ifr) == -1){
      if(errno == EINVAL) continue;
      log_errno("ioctl");
      log_err("failed to get interface broadcast address");
      continue;
    }

    if(*brd_n >= brd_max){
      log_err("max interfaces reached");
      break;
    }

    sa = (struct sockaddr_in *)&ifr.ifr_broadaddr;
    brd[(*brd_n)++] = host32be(sa->sin_addr.s_addr);
  }

  close(fd);
  return 1;

close:
  close(fd);
fail:
  return 0;
}

