#ifndef PAK_H
#define PAK_H

#include "peer/key.h"
#include <config.h> //GREASYPAK_MTU
#include <crypt/crypt.h>
#include <net/ipv4.h>
#include <net/udp.h>
#include <util/util.h>
#include <stdint.h>
#include <stdbool.h>

#define PAK_IPV4_MTU            GREASYPAK_MTU
#define PAK_SIG_SIZE            HMAC384_SIZE
#define PAK_CIPH_BLOCK          SERPENT_BLOCKSIZE
#define PAK_CIPH_FLOOR(S)       ((S)/PAK_CIPH_BLOCK*PAK_CIPH_BLOCK)
#define PAK_CIPH_CEIL(S)        (CEIL_DIV((S),PAK_CIPH_BLOCK)*PAK_CIPH_BLOCK)
#define PAK_CIPH_VALID(S)       ((S)%PAK_CIPH_BLOCK == 0)
#define PAK_SIGKEY_SIZE         32
#define PAK_CIPHKEY_SIZE        32

#define PAK_PROTO               0xFF

#define PAK_TYPE_IP             0x00
#define PAK_TYPE_FLOOD          0x01
#define PAK_TYPE_PROD           0x02
#define PAK_TYPE_ADDR           0x03
#define PAK_TYPE_KEYOFFER1      0x04
#define PAK_TYPE_KEYOFFER2      0x05
#define PAK_TYPE_KEYSLICE1      0x06
#define PAK_TYPE_KEYSLICE2      0x07
#define PAK_TYPE_KEYDONE1       0x08
#define PAK_TYPE_KEYDONE2       0x09

////////////////////////////////////////////////////////////////////////////////

#define BLK_MAXSIZE_NOM         (PAK_IPV4_MTU-IPV4_HDR_MAXSIZE-UDP_HDR_SIZE)
#define BLK_MAXSIZE             (BLK_RED_OFF+RED_MAXSIZE)
#define BLK_MINSIZE             (BLK_RED_OFF+RED_MINSIZE)
#define BLK_SIZE(R)             (BLK_RED_OFF+(R))
#define BLK_VALIDSIZE(B)        (((B) >= BLK_MINSIZE) &&\
                                 ((B) <= BLK_MAXSIZE) &&\
                                 PAK_CIPH_VALID(RED_SIZEB(B)))
#define BLK_SIG_OFF             0
#define BLK_SIG_SIZE            PAK_SIG_SIZE
#define BLK_RED_OFF             (BLK_SIG_OFF+BLK_SIG_SIZE)

////////////////////////////////////////////////////////////////////////////////

#define RED_MAXSIZE             PAK_CIPH_FLOOR(BLK_MAXSIZE_NOM-BLK_RED_OFF)
#define RED_MINSIZE             RED_SIZE(0)
#define RED_SIZE(L)             PAK_CIPH_CEIL(RED_LOAD_OFF+(L))
#define RED_SIZEB(B)            ((B)-BLK_RED_OFF)
#define RED_VALIDSIZE(R)        (((R) >= RED_MINSIZE) &&\
                                 ((R) <= RED_MAXSIZE) &&\
                                 PAK_CIPH_VALID(R))
#define RED_PADDING(R,L)        ((R)-(RED_LOAD_OFF+(L)))
#define RED_NONCE_OFF           0
#define RED_NONCE_SIZE          PAK_CIPH_BLOCK
#define RED_SEQNO_OFF           (RED_NONCE_OFF+RED_NONCE_SIZE)
#define RED_SEQNO_SIZE          sizeof(uint32_t)
#define RED_PROTO_OFF           (RED_SEQNO_OFF+RED_SEQNO_SIZE)
#define RED_PROTO_SIZE          sizeof(uint8_t)
#define RED_TYPE_OFF            (RED_PROTO_OFF+RED_PROTO_SIZE)
#define RED_TYPE_SIZE           sizeof(uint8_t)
#define RED_PAD_OFF             (RED_TYPE_OFF+RED_TYPE_SIZE)
#define RED_PAD_SIZE            sizeof(uint16_t)
#define RED_LOAD_OFF            (RED_PAD_OFF+RED_PAD_SIZE)
#define RED_LOAD_MAXSIZE        (RED_MAXSIZE-RED_LOAD_OFF)
#define RED_LOAD_SIZE(R)        ((R)-RED_LOAD_OFF)
#define RED_LOAD_VALIDSIZE(L)   (((size_t)(L) <= RED_LOAD_MAXSIZE))

////////////////////////////////////////////////////////////////////////////////

#define IP_PAK_OFF              0
#define IP_PAK_MAXSIZE          RED_LOAD_MAXSIZE

////////////////////////////////////////////////////////////////////////////////

#define FLOOD_SIZE(L)           (FLOOD_LOAD_OFF+(L))
#define FLOOD_MINSIZE           FLOOD_SIZE(0)
#define FLOOD_BUFSIZE           FLOOD_SIZE(BLK_MAXSIZE)
#define FLOOD_SRC_OFF           0
#define FLOOD_SRC_SIZE          sizeof(uint32_t)
#define FLOOD_DST_OFF           (FLOOD_SRC_OFF+FLOOD_SRC_SIZE)
#define FLOOD_DST_SIZE          sizeof(uint32_t)
#define FLOOD_SIZE_OFF          (FLOOD_DST_OFF+FLOOD_DST_SIZE)
#define FLOOD_SIZE_SIZE         sizeof(uint16_t)
#define FLOOD_TTL_OFF           (FLOOD_SIZE_OFF+FLOOD_SIZE_SIZE)
#define FLOOD_TTL_SIZE          sizeof(uint8_t)
#define FLOOD_PAD_OFF           (FLOOD_TTL_OFF+FLOOD_TTL_SIZE)
#define FLOOD_PAD_SIZE          sizeof(uint8_t)
#define FLOOD_LOAD_OFF          (FLOOD_PAD_OFF+FLOOD_PAD_SIZE)
#define FLOOD_LOAD_MAXSIZE      (RED_LOAD_MAXSIZE-FLOOD_LOAD_OFF)
#define FLOOD_LOAD_SIZE(L)      ((L)-FLOOD_LOAD_OFF)
#define FLOOD_LOAD_VALIDSIZE(L) (((size_t)(L) <= FLOOD_LOAD_MAXSIZE))

////////////////////////////////////////////////////////////////////////////////

#define PROD_FLAGS_OFF          0
#define PROD_FLAGS_SIZE         sizeof(uint32_t)
#define PROD_SEQNO_OFF          (PROD_FLAGS_OFF+PROD_FLAGS_SIZE)
#define PROD_SEQNO_SIZE         sizeof(uint32_t)
#define PROD_IP_OFF             (PROD_SEQNO_OFF+PROD_SEQNO_SIZE)
#define PROD_IP_SIZE            sizeof(uint32_t)
#define PROD_PORT_OFF           (PROD_IP_OFF+PROD_IP_SIZE)
#define PROD_PORT_SIZE          sizeof(uint16_t)
#define PROD_SIZE               (PROD_PORT_OFF+PROD_PORT_SIZE)

#define PROD_REQANS             0x00000001

////////////////////////////////////////////////////////////////////////////////

#define ADDR_IP_OFF             0
#define ADDR_IP_SIZE            sizeof(uint32_t)
#define ADDR_PORT_OFF           (ADDR_IP_OFF+ADDR_IP_SIZE)
#define ADDR_PORT_SIZE          sizeof(uint16_t)
#define ADDR_SIZE               (ADDR_PORT_OFF+ADDR_PORT_SIZE)

////////////////////////////////////////////////////////////////////////////////

static void pak_h_assert(void){
  ASSERT(RED_MINSIZE      <= RED_MAXSIZE);
  ASSERT(IPV4_HDR_MAXSIZE <= RED_LOAD_MAXSIZE);
  ASSERT(PROD_SIZE        <= RED_LOAD_MAXSIZE);
  ASSERT(ADDR_SIZE        <= RED_LOAD_MAXSIZE);
  ASSERT(ADDR_SIZE        <= FLOOD_LOAD_MAXSIZE);
}

bool red_hdr      (uint8_t         *r,
                   uint8_t          type,
                   pkey_t const    *pk);

void load_copy    (uint8_t         *l,
                   uint8_t const   *load,
                   size_t           load_size);

void blk_ciph     (uint8_t         *b,
                   uint8_t const   *r,
                   size_t           r_size,
                   uint8_t const   *key);

void blk_deciph   (uint8_t         *r,
                   uint8_t const   *b,
                   size_t           r_size,
                   uint8_t const   *key);

void blk_sign     (uint8_t         *b,
                   size_t           r_size,
                   uint8_t const   *key);

bool blk_verify   (uint8_t const   *b,
                   size_t           r_size,
                   uint8_t const   *key);

bool blk_prepare  (uint8_t         *b,
                   size_t           load_size,
                   uint8_t          type,
                   pkey_t const    *pk,
                   size_t          *b_size_out);

#endif

