#include "key.h"
#include <string.h>
#include <crypt/crypt.h>
#include <util/util.h>
#include <util/serial.h>
#include <util/hex.h>

extern uint32_t     pkey_sent   (pkey_t *pk);
extern bool         pkey_self   (pkey_t *pk,uint32_t seq);
extern uint32_t     pkey_txnext (pkey_t const *pk);
extern bool         pkey_txhalf (pkey_t const *pk);
extern bool         pkey_txlim  (pkey_t const *pk);

void pkey_init(pkey_t *pk){
  memset_ffs(pk->key.full, 0x00,sizeof(pk->key.full));
  memset_ffs(pk->ciph_serp,0x00,sizeof(pk->ciph_serp));
  pkey_rzero(pk);
  pk->rxseq = 0;
  pk->txseq = 0;
  pk->stamp = 0;
}

void pkey_prep(pkey_t *pk){
  serpent_init(pk->ciph_serp,
               pk->key.sub.ciph,
               sizeof(pk->key.sub.ciph));
}

void pkey_set(pkey_t *pk,uint8_t const *key){
  memcpy(pk->key.full,key,sizeof(pk->key.full));
  pkey_prep(pk);
}

void pkey_rzero(pkey_t *pk){
  memset(pk->rxbuf,0x00,sizeof(pk->rxbuf));
}

static inline size_t div_8(size_t x){
  return x>>3;
}

static inline size_t mod_8(size_t x){
  return x & 8-1;
}

static inline size_t mod_rxbuf_bits(size_t i){
  return i & RXBUF_SIZE*8-1;
}

static bool bit_get(pkey_t const *pk,size_t i){
  i = mod_rxbuf_bits(i);
  return pk->rxbuf[div_8(i)]>>mod_8(i) & 0x01;
}

static void bit_set(pkey_t *pk,size_t i){
  i = mod_rxbuf_bits(i);
  pk->rxbuf[div_8(i)] |= 0x01<<mod_8(i);
}

static void bit_clr(pkey_t *pk,size_t i){
  i = mod_rxbuf_bits(i);
  pk->rxbuf[div_8(i)] &= ~(0x01<<mod_8(i));
}

bool pkey_recv(pkey_t *pk,uint32_t rxseq){
  if(rxseq > pk->rxseq){
    if(rxseq-pk->rxseq >= RXBUF_SIZE*8-1){
      pkey_rzero(pk);
    }else for(size_t x = pk->rxseq+1;x < rxseq;x++){
      bit_clr(pk,x);
    }
    bit_set(pk,rxseq);
    pk->rxseq = rxseq;
    return 1;
  }else if(pk->rxseq-rxseq >= RXBUF_SIZE*8){
    return 0;
  }else if(bit_get(pk,rxseq)){
    return 0;
  }else{
    bit_set(pk,rxseq);
    return 1;
  }
}

bool key_equal(uint8_t const *a,uint8_t const *b){
  return memcmp(a,b,KEY_SIZE) == 0;
}

void key_copy(uint8_t *dst,uint8_t const *src){
  memcpy(dst,src,KEY_SIZE);
}

void key_xor(uint8_t *dst,uint8_t const *a,uint8_t const *b){
  for(size_t x = 0;x < KEY_SIZE;x++){
    dst[x] = a[x] ^ b[x];
  }
}

uint32_t key_crc(uint8_t const *k){
  uint8_t sha[SHA512_SIZE];
  sha512(sha,k,KEY_SIZE);
  return crc32(sha,sizeof(sha),0);
}

void key_crc_hex(char *buf,size_t buf_size,uint8_t const *k){
  uint32_t  crc = host32be(key_crc(k));
  uint8_t   crc_buf[sizeof(crc)];
  memcpy(crc_buf,&crc,sizeof(crc));
  hex_enc(buf,buf_size,crc_buf,sizeof(crc_buf));
}

