#include "peer.h"
#include "default.h"
#include <io/fs.h>
#include <io/sx.h>
#include <net/util.h>
#include <daemon/internal.h>
#include <daemon/type/type.h>
#include <util/util.h>
#include <util/time.h>
#include <util/string.h>
#include <util/hex.h>
#include <log/log.h>

#include <string.h>
#include <dirent.h>
#include <unistd.h>

#define SPEC_STRUCT p
#define SPEC_FAIL   spec_fail

peer_t peer[PEER_MAX];
size_t peer_n;

void peer_init(peer_t *p){
  p->id[0]        = 0;
  p->ip           = DEF_IP;
  p->locaddr      = DEF_LOCADDR;
  p->extaddr      = DEF_EXTADDR;
  BUF_ZERO(p->pkey);
  p->pkey_n       = 0;
  p->flags        = DEF_FLAGS;

  p->grx_n        = DEF_GRXN;
  p->grx_b        = DEF_GRXB;
  p->gtx_n        = DEF_GTXN;
  p->gtx_b        = DEF_GTXB;
  p->nrx_n        = DEF_NRXN;
  p->nrx_b        = DEF_NRXB;
  p->ntx_n        = DEF_NTXN;
  p->ntx_b        = DEF_NTXB;
  p->rrx_n        = DEF_RRXN;
  p->rrx_b        = DEF_RRXB;
  p->rtx_n        = DEF_RTXN;
  p->rtx_b        = DEF_RTXB;
  p->drx_n        = DEF_DRXN;
  p->drx_b        = DEF_DRXB;
  p->dtx_n        = DEF_DTXN;
  p->dtx_b        = DEF_DTXB;

  p->edit         = 0;

  p->ntime        = 0;
  p->prod_ntime   = 0;
  p->prod_txntime = 0;
  p->prod_txseq   = 0;
  p->prod_rxseq   = 0;
  p->addr_ntime   = 0;
  rekey_init(&p->rekey);
}

static void peer_path(char *buf,size_t buf_size,char const *id){
  snprintf(buf,buf_size,"%s/%s",PEER_PATH,id);
}

//must not be caled on peer alredy in the peer aray. cuz of lookups
static bool peer_load(peer_t *p,char const *id){
  char          path[PATH_BUF];
  val_t         sx;
  lisp_frame_t  frame;
  char const   *err;
  char const   *err_opt;

  if(!id_valid(id,&err)){
    log_err("invalid peer id %s, %s",id,err);
    return 0;
  }

  if(peer_lookup(id)){
    log_err("peer id %s already in use",id);
    return 0;
  }

  peer_init(p);
  strzcpy(p->id,id,sizeof(p->id));

  lisp_store(&frame);
  peer_path(path,sizeof(path),p->id);
  if(!sx_load(path,&sx)) goto pop;

  lisp_try{
#include <daemon/type/spec_set.h>
#include "spec.h"
  }lisp_catch{
    log_lisperr();
    goto pop;
  }

  lisp_load(&frame);

  if(!ip_valid(p->ip,&err)){
    char ip_buf[IPV4_BUF];
    ipv4_fmt(ip_buf,sizeof(ip_buf),p->ip);
    log_err("peer ip %s %s",ip_buf,err);
    goto clear;
  }

  p->edit = 0;
  return peer_save(p);

spec_fail:
  log_err("option %s: %s",err_opt,err);
pop:
  lisp_load(&frame);
clear:
  peer_init(p);
  log_err("failed to load peer file %s",path);
  return 0;
}

bool peer_save(peer_t *p){
  char          path[PATH_BUF];
  val_t         sx;
  lisp_frame_t  frame;

  lisp_store(&frame);
  lisp_try{
    sx = list(PP_ALIST);
#include <daemon/type/spec_get.h>
#include "spec.h"
    sx = nreverse(sx);
  }lisp_catch{
    log_lisperr();
    goto pop;
  }

  peer_path(path,sizeof(path),p->id);
  if(!sx_save(path,sx)) goto pop;
  lisp_load(&frame);
  p->edit = 0;
  return 1;

pop:
  lisp_load(&frame);
  log_err("failed to save peer file %s",path);
  return 0;
}

void peer_edit(peer_t *p){
  p->edit = 1;
}

bool peer_flush(peer_t *p){
  if(!p->edit) return 1;
  return peer_save(p);
}

peer_t *peer_alloc(void){
  if(peer_n >= PEER_MAX){
    return 0;
  }
  return peer+peer_n;
}

void peer_add(void){
  assert(peer_n < PEER_MAX);
  peer_n++;
}

bool peer_del(peer_t *p){
  char path[PATH_BUF];

  peer_path(path,sizeof(path),p->id);
  if(!fs_zero(path)) return 0;
  if(!fs_rm_f(path)) return 0;

  //TODO: this is a hak update per bit arays imsted
  flood_flush();

  peer_init(p);
  assert(peer_n > 0);
  ARR_DEL(peer,peer_n,p);
  peer_n--;
  return 1;
}

bool peer_rename(peer_t *p,char const *id,char const **err){
  char path_old[PATH_BUF];
  char path_new[PATH_BUF];

  if(peer_lookup(id)){
    if(err) *err = "id already in use";
    return 0;
  }

  peer_path(path_old,sizeof(path_old),p->id);
  peer_path(path_new,sizeof(path_new),id);

  if(!fs_rename(path_old,path_new)){
    if(err) *err = "failed to rename peer file";
    return 0;
  }

  return 1;
}

bool peer_key_add(peer_t *p,uint8_t const *key,uint32_t ip_ref){
  char    hex[HEX_BUF(CRC32_SIZE)];
  pkey_t *pk;

  if(p->pkey_n >= KEY_MAX){
    return 0;
  }

  pk = &p->pkey[p->pkey_n++];
  pkey_init(pk);
  pkey_set(pk,key);

  if(ip_ref < p->ip){
    pk->rxseq = 0;
    pk->txseq = PKEY_SEQBIT;
  }else{
    pk->rxseq = PKEY_SEQBIT;
    pk->txseq = 0;
  }

  peer_edit(p);
  key_crc_hex(hex,sizeof(hex),key);
  log_out("added key %s for peer %s",hex,p->id);
  return 1;
}

void peer_pkey_del(peer_t *p,pkey_t *pk){
  char hex[HEX_BUF(CRC32_SIZE)];
  key_crc_hex(hex,sizeof(hex),pk->key.full);

  pkey_init(pk);
  ARR_DEL(p->pkey,p->pkey_n,pk);
  p->pkey_n--;
  peer_edit(p);

  log_out("deleted key %s for peer %s",hex,p->id);
}

bool peer_key_del(peer_t *p,uint8_t const *key){
  pkey_t *pk = peer_key_lookup(p,key);

  if(!pk){
    char hex[HEX_BUF(CRC32_SIZE)];
    key_crc_hex(hex,sizeof(hex),key);
    log_err("key %s doesn't exist for peer %s, cannot delete",
            hex,p->id);
    return 0;
  }

  peer_pkey_del(p,pk);
  return 1;
}

pkey_t *peer_key_lookup(peer_t const *p,uint8_t const *key){
  for(size_t x = 0;x < p->pkey_n;x++){
    pkey_t const *c = p->pkey+x;
    if(key_equal(c->key.full,key)){
      return PTR_UNCONST(c);
    }
  }
  return 0;
}

pkey_t *peer_key_lookupc(peer_t const *p,uint32_t crc){
  for(size_t x = 0;x < p->pkey_n;x++){
    pkey_t const *c = p->pkey+x;
    if(crc == key_crc(c->key.full)){
      return PTR_UNCONST(c);
    }
  }
  return 0;
}

//TODO: store ptr shit imsted vhen chagne
pkey_t *peer_key_recent(peer_t const *p){
  uint64_t      max_stamp = 0;
  pkey_t const *max_key   = 0;
  for(size_t x = 0;x < p->pkey_n;x++){
    pkey_t const *c = p->pkey+x;
    if(c->stamp < max_stamp) continue;
    max_stamp = c->stamp;
    max_key   = c;
  }
  return PTR_UNCONST(max_key);
}

peer_t *peer_lookup(char const *id){
  for(size_t x = 0;x < peer_n;x++){
    peer_t *p = peer+x;
    if(strcmp(p->id,id) == 0) return p;
  }
  return 0;
}

peer_t *peer_lookup_ip(uint32_t ip){
  for(size_t x = 0;x < peer_n;x++){
    peer_t *p = peer+x;
    if(p->ip == ip) return p;
  }
  return 0;
}

peer_t *peer_lookup_key(uint8_t const *key){
  for(size_t x = 0;x < peer_n;x++){
    peer_t *p = peer+x;
    if(peer_key_lookup(p,key)) return p;
  }
  return 0;
}

void peer_addr_set(peer_t *p,addr_t addr,bool local){
  addr_t *prev  = local ? &p->locaddr : &p->extaddr;
  char    ips_a[IPV4_BUF];
  char    ips_b[IPV4_BUF];

  if(!addr_equal(addr,*prev)){
    log_out("peer %s %s address %s:%u -> %s:%u",
            p->id,
            local ? "local" : "external",
            IP_A(prev->ip),
            prev->port,
            IP_B(addr.ip),
            addr.port);
  }

  *prev = addr;
  peer_edit(p);
}

void peer_addr_auto(peer_t *p,addr_t addr){
  peer_addr_set(p,addr,ipv4_local(addr.ip));
}

addr_t *peer_effaddr(peer_t const *p,uint64_t time){
  bool          loc_null = addr_null(p->locaddr);
  bool          ext_null = addr_null(p->extaddr);
  addr_t const *addr;

  if(loc_null){
    addr = ext_null ? 0 : &p->extaddr;
  }else if(ext_null){
    addr = &p->locaddr;
  }else if(time < p->locaddr.stamp+COLD_TIMEOUT){
    addr = &p->locaddr;
  }else{
    addr = &p->extaddr;
  }

  return PTR_UNCONST(addr);
}

bool peer_cancrypt(peer_t const  *p,
                   pkey_t       **pk,
                   char const   **err){

  pkey_t *_pk = peer_key_recent(p);
  if(!_pk){
    if(err) *err = "no keys configured";
    return 0;
  }
    
  if(pkey_txlim(_pk)){
    if(err) *err = "tx sequence number exhausted";
    return 0;
  }

  if(pk) *pk = _pk;
  return 1;
}

bool peer_cansend(peer_t const  *p,
                  pkey_t       **pk,
                  addr_t       **addr,
                  char const   **err){

  addr_t *_addr;

  if(!peer_cancrypt(p,pk,err)){
    return 0;
  }

  _addr = peer_effaddr(p,time_sec());
  if(!_addr){
    if(err) *err = "no address configured";
    return 0;
  }

  if(p->flags & PEER_MUTE){
    if(err) *err = "mute flag set";
    return 0;
  }

  if(addr) *addr = _addr;
  return 1;
}

bool peer_canbrd(peer_t const  *p,
                 pkey_t       **pk,
                 char const   **err){

  if(!peer_cancrypt(p,pk,err)){
    return 0;
  }

  if(p->locaddr.port == 0){
    if(err) *err = "local address port not configured";
    return 0;
  }

  return 1;
}

bool peer_cold(peer_t *p,uint64_t time){
  uint64_t stamp = p->extaddr.stamp > p->locaddr.stamp ?
                   p->extaddr.stamp : p->locaddr.stamp;
  return time >= stamp+COLD_TIMEOUT;
}

bool peer_cold_loc(peer_t *p,uint64_t time){
  return time >= p->locaddr.stamp+COLD_TIMEOUT;
}

size_t peer_idx(peer_t const *p){
  return ARR_INDEXOF(peer,p);
}

//hak so wraps to 0 after 600 years imsted of error when store to disk
static int64_t wrap_to_int64(uint64_t i){
  return i & INT64_MAX;
}

#define STATS_IMPL(N)\
  void peer_##N(peer_t *p,uint64_t b){\
    p->N##_n = wrap_to_int64(p->N##_n+1);\
    p->N##_b = wrap_to_int64(p->N##_b+b);\
    peer_edit(p);\
  }

STATS_IMPL(grx)
STATS_IMPL(gtx)
STATS_IMPL(nrx)
STATS_IMPL(ntx)
STATS_IMPL(rrx)
STATS_IMPL(rtx)
STATS_IMPL(drx)
STATS_IMPL(dtx)

void peer_init_all(void){
  for(size_t x = 0;x < peer_n;x++){
    peer_t *p = peer+x;
    peer_init(p);
  }
}

bool peer_load_all(void){
  DIR            *d;
  struct dirent  *de;

  peer_n = 0;

  if((d = opendir(PEER_PATH)) == 0){
    log_errno("opendir");
    log_err("failed to open directory for listing %s",PEER_PATH);
    return 0;
  }

  while(de = readdir(d)){
    peer_t *p;

    if(strcmp(de->d_name,"." ) == 0) continue;
    if(strcmp(de->d_name,"..") == 0) continue;

    p = peer_alloc();
    if(!p){
      log_err("too many peers, increase PEER_MAX");
      goto fail;
    }

    if(!peer_load(p,de->d_name)) goto fail;
    peer_add();
  }

  closedir(d);
  return 1;

fail:
  closedir(d);
  return 0;
}

bool peer_flush_all(void){
  bool r = 1;
  for(size_t x = 0;x < peer_n;x++){
    peer_t *p = peer+x;
    if(!peer_flush(p)) r = 0;
  }
  return r;
}

