#include "internal.h"
#include "pak.h"
#include "peer/peer.h"
#include "state/state.h"
#include <util/util.h>
#include <util/time.h>
#include <log/log.h>

#include <string.h>
#include <assert.h>

#define STATE_INACTIVE      WAIT_KEYOFFER1

enum{
  WAIT_KEYOFFER1,
  WAIT_KEYOFFER2,
  WAIT_KEYSLICE1,
  WAIT_KEYSLICE2,
  WAIT_KEYDONE1,
  WAIT_KEYDONE2,
  WAIT_PACKET1,
};

void rekey_init(rekey_t *rk){
  rk->state = STATE_INACTIVE;
  rk->ntime = 0;
  rk->nlast = 0;
  BUF_ZERO(rk->offer_self);
  BUF_ZERO(rk->offer_peer);
  BUF_ZERO(rk->slice_self);
  BUF_ZERO(rk->slice_peer);
  BUF_ZERO(rk->key_old);
  BUF_ZERO(rk->key_new);
}

bool rekey_active(rekey_t *rk){
  return rk->state != STATE_INACTIVE;
}

bool rekey_keyexists(rekey_t *rk){
  return rk->state == WAIT_KEYDONE1 ||
         rk->state == WAIT_KEYDONE2 ||
         rk->state == WAIT_PACKET1;
}

static void rekey_abort(peer_t *p,char const *reason){
  rekey_t *rk = &p->rekey;

  if(rekey_keyexists(rk)){
    peer_key_del(p,rk->key_new);
  }

  log_err("rekeying with peer %s aborted: %s",p->id,reason);
  rekey_init(rk);
}

static bool keyoffer_gen(peer_t *p){
  rekey_t *rk = &p->rekey;

  if(!key_gen(rk->slice_self)){
    rekey_abort(p,"failed to generate key slice");
    return 0;
  }
  sha512(rk->offer_self,rk->slice_self,sizeof(rk->slice_self));

  return 1;
}

static bool keyoffer_send(peer_t *p,bool first){
  rekey_t    *rk = &p->rekey;
  pkey_t     *pk;
  addr_t     *addr;
  char const *err;

  if(!peer_cansend(p,&pk,&addr,&err)){
    rekey_abort(p,err);
    return 0;
  }

  if(!tx_load(rk->offer_self,
              sizeof(rk->offer_self),
              first ? PAK_TYPE_KEYOFFER1 : PAK_TYPE_KEYOFFER2,
              p,
              pk,
              addr_to_in(*addr))){
    rekey_abort(p,"failed to send key offer");
    return 0;
  }

  rk->nlast = mono_nsec();
  return 1;
}

static bool keyslice_send(peer_t *p,bool first){
  rekey_t    *rk = &p->rekey;
  pkey_t     *pk;
  addr_t     *addr;
  char const *err;

  if(!peer_cansend(p,&pk,&addr,&err)){
    rekey_abort(p,err);
    return 0;
  }

  if(!tx_load(rk->slice_self,
              sizeof(rk->slice_self),
              first ? PAK_TYPE_KEYSLICE1 : PAK_TYPE_KEYSLICE2,
              p,
              pk,
              addr_to_in(*addr))){
    rekey_abort(p,"failed to send key slice");
    return 0;
  }

  rk->nlast = mono_nsec();
  return 1;
}

static bool keyoffer_recv(peer_t *p,uint8_t const *offer,size_t size){
  rekey_t *rk = &p->rekey;

  if(size < sizeof(rk->offer_peer)){
    rekey_abort(p,"invalid key offer size");
    return 0;
  }

  memcpy(rk->offer_peer,offer,sizeof(rk->offer_peer));

  if(memcmp(rk->offer_peer,
            rk->offer_self,
            sizeof(rk->offer_peer)) == 0){
    rekey_abort(p,"received key offer was "
                  "identical to key offer sent");
    return 0;
  }

  return 1;
}

static bool keyslice_recv(peer_t *p,uint8_t *slice,size_t size){
  uint8_t   hash[SHA512_SIZE];
  rekey_t  *rk = &p->rekey;
  pkey_t   *pk;

  if(size < sizeof(rk->slice_peer)){
    rekey_abort(p,"invalid key slice size");
    return 0;
  }

  key_copy(rk->slice_peer,slice);
  memset_ffs(slice,0x00,KEY_SIZE);
  sha512(hash,rk->slice_peer,sizeof(rk->slice_peer));

  if(memcmp(hash,rk->offer_peer,sizeof(hash)) != 0){
    BUF_ZERO(hash);
    rekey_abort(p,"received key slice did not match key offer");
    return 0;
  }
  BUF_ZERO(hash);

  if(entropy_missing(rk->slice_peer,sizeof(rk->slice_peer))){
    rekey_abort(p,"entropy obviously missing "
                  "from received key slice");
    return 0;
  }

  pk = peer_key_recent(p);
  assert(pk);

  key_copy(rk->key_old,pk->key.full);
  key_xor(rk->key_new,rk->key_old,rk->slice_self);
  key_xor(rk->key_new,rk->key_new,rk->slice_peer);

  if(!peer_key_add(p,rk->key_new,state.ip)){
    rekey_abort(p,"failed to add new key");
    return 0;
  }

  if(!peer_save(p)){
    rekey_abort(p,"failed to save peer");
    return 0;
  }

  return 1;
}

static pkey_t *rekey_pkey(peer_t *p){
  rekey_t  *rk = &p->rekey;
  pkey_t   *pk = peer_key_lookup(p,rk->key_new);

  if(!pk){
    rekey_abort(p,"new key has been deleted");
    return 0;
  }

  return pk;
}

static bool keydone_recv(peer_t *p,pkey_t *pk){
  rekey_t *rk = &p->rekey;

  if(!key_equal(pk->key.full,rk->key_new)){
    return 0;
  }

  log_out("rekeying with peer %s successful",p->id);
  return 1;
}

static bool keydone_send(peer_t *p,bool first){
  rekey_t    *rk = &p->rekey;
  pkey_t     *pk = rekey_pkey(p);
  addr_t     *addr;
  char const *err;

  if(!pk) return 0;

  if(!peer_cansend(p,0,&addr,&err)){
    rekey_abort(p,err);
    return 0;
  }

  if(!tx_load(0,
              0,
              first ? PAK_TYPE_KEYDONE1 : PAK_TYPE_KEYDONE2,
              p,
              pk,
              addr_to_in(*addr))){
    rekey_abort(p,"failed to send packet");
    return 0;
  }

  rk->nlast = mono_nsec();
  return 1;
}

static bool initiate(peer_t *p){
  rekey_t    *rk = &p->rekey;
  char const *err;

  if(!peer_cansend(p,0,0,&err)){
    log_err("%s for peer %s, cannot rekey",err,p->id);
    return 0;
  }

  if(rekey_active(rk)){
    log_err("rekeying with peer %s already in progress",p->id);
    return 0;
  }

  rk->state = WAIT_KEYOFFER2;
  rk->ntime = mono_nsec();

  if(!keyoffer_gen(p)) return 0;
  if(!keyoffer_send(p,1)) return 0;

  log_out("rekeying with peer %s initiated",p->id);
  return 1;
}

bool rekey_initiate(peer_t *p){
  if(p){
    return initiate(p);
  }else{
    bool r = 1;
    for(size_t x = 0;x < peer_n;x++){
      peer_t *q = peer+x;
      if(!peer_cansend(q,0,0,0))  continue; //so dont err
      if(rekey_active(&q->rekey)) continue;
      if(!initiate(q)) r = 0;
    }
    return r;
  }
}

bool rekey_recv(peer_t *p,pkey_t *pk,uint8_t type,uint8_t *l,size_t l_size){
  rekey_t *rk = &p->rekey;

  switch(type){
  case PAK_TYPE_KEYOFFER1:
    if(rk->state != WAIT_KEYOFFER1) break;
    if(!keyoffer_recv(p,l,l_size))  break;
    log_out("rekeying offer from peer %s accepted",p->id);
    if(!keyoffer_gen(p))            break;
    if(!keyoffer_send(p,0))         break;
    rk->state = WAIT_KEYSLICE1;
    rk->ntime = mono_nsec();
    return 1;

  case PAK_TYPE_KEYOFFER2:
    if(rk->state != WAIT_KEYOFFER2) break;
    if(!keyoffer_recv(p,l,l_size))  break;
    if(!keyslice_send(p,1))         break;
    rk->state = WAIT_KEYSLICE2;
    return 1;

  case PAK_TYPE_KEYSLICE1:
    if(rk->state != WAIT_KEYSLICE1) break;
    if(!keyslice_recv(p,l,l_size))  break;
    if(!keyslice_send(p,0))         break;
    rk->state = WAIT_KEYDONE1;
    return 1;

  case PAK_TYPE_KEYSLICE2:
    if(rk->state != WAIT_KEYSLICE2) break;
    if(!keyslice_recv(p,l,l_size))  break;
    if(!keydone_send(p,1))          break;
    rk->state = WAIT_KEYDONE2;
    return 1;

  case PAK_TYPE_KEYDONE1:
    if(rk->state != WAIT_KEYDONE1)  break;
    if(!keydone_recv(p,pk))         break;
    if(!keydone_send(p,0))          break;
    rk->state = WAIT_PACKET1;
    return 1;

  case PAK_TYPE_KEYDONE2:
    if(rk->state != WAIT_KEYDONE2)  break;
    if(!keydone_recv(p,pk))         break;
    prod_send(p,0);
    peer_key_del(p,rk->key_old);
    rekey_init(rk);
    return 1;

  default:
    if(rk->state != WAIT_PACKET1)   break;
    if(!key_equal(pk->key.full,
                  rk->key_new))     break;
    peer_key_del(p,rk->key_old);
    rekey_init(rk);
    return 1;
  }

  return 0;
}

static bool rekey_retransmit(peer_t *p,uint64_t ntime){
  rekey_t *rk = &p->rekey;

  if(ntime < rk->nlast+REKEY_RETRY) return 1;
  if(!peer_cansend(p,0,0,0))        return 0;

  switch(rk->state){
  case WAIT_KEYOFFER2:  return keyoffer_send (p,1);
  case WAIT_KEYSLICE1:  return keyoffer_send (p,0);
  case WAIT_KEYSLICE2:  return keyslice_send (p,1);
  case WAIT_KEYDONE1:   return keyslice_send (p,0);
  case WAIT_KEYDONE2:   return keydone_send  (p,1);
  case WAIT_PACKET1:    return keydone_send  (p,0);
  default:              return 1;
  }
}

void rekey_tick(uint64_t ntime){
  for(size_t x = 0;x < peer_n;x++){
    peer_t   *p   = peer+x;
    rekey_t  *rk  = &p->rekey;

    if(!rekey_active(rk))               continue;
    rekey_retransmit(p,ntime);
    if(rk->state == WAIT_PACKET1)       continue;
    if(ntime < rk->ntime+REKEY_TIMEOUT) continue;

    rekey_init(rk);
    log_out("rekeying with peer %s timed out",p->id);
  }
}

