#include "internal.h"
#include "pak.h"
#include <config.h> //MEMSET_RED, LOG_NETSPAM
#include "state/state.h"
#include "peer/peer.h"
#include "peer/key.h"
#include <util/time.h>
#include <log/log.h>

static bool ip_recv(peer_t *p,uint8_t *l,size_t l_size){
  uint8_t  *ipak = l+IP_PAK_OFF;
  size_t    ipak_size;
  uint32_t  src;
  uint32_t  dst;
  char      ips_a[IPV4_BUF];
  char      ips_b[IPV4_BUF];

  if(l_size < IPV4_HDR_MINSIZE){
    log_err("incoming ip packet invalid load size (%u)",l_size);
    return 0;
  }

  if(!ver_check(ipak,DIR_IN)) return 0;

  ipak_size = ipv4_len_get(ipak);
  if(ipak_size > l_size){
    log_err("incoming ipv4 length exceeds load size (%zu > %zu)",
            ipak_size,l_size);
    return 0;
  }

  if(!sum_check (ipak,DIR_IN)) return 0;
  if(!ttl_dec   (ipak,DIR_IN)) return 0;

  src = ipv4_src_get(ipak);
  dst = ipv4_dst_get(ipak);

  if(src != p->ip){
    log_err("incoming source ip mismatch (%s != %s)",
            IP_A(src),IP_B(p->ip));
    return 0;
  }

  if(dst != state.ip){
    log_err("incoming destination ip mismatch (%s != %s)",
            IP_A(dst),IP_B(state.ip));
    return 0;
  }

  ipv4_update_sum(ipak);
  if(tun_write(ipak,ipak_size)){
    state_nrx(ipak_size);
    peer_nrx(p,ipak_size);
  }else{
    state_drx(ipak_size);
    peer_drx(p,ipak_size);
  }
  return 1;
}

static bool find_key(uint8_t *b,size_t r_size,peer_t **p_out,pkey_t **pk_out){
  for(size_t x = 0;x < peer_n;x++){
    peer_t *p = peer+x;
    for(size_t y = 0;y < p->pkey_n;y++){
      pkey_t *pk = p->pkey+y;
      if(!blk_verify(b,r_size,pk->key.sub.sign)) continue;
      *p_out  = p;
      *pk_out = pk;
      return 1;
    }
  }
  return 0;
}

bool rx_blk(uint8_t *b,size_t b_size,addr_in_t addr_in,bool matryoshka){
  peer_t   *p = 0;
  pkey_t   *pk;
  addr_t    addr;
  uint8_t  *r;
  uint8_t  *l;
  size_t    r_size;
  size_t    l_size;
  uint32_t  rxseq;
  uint8_t   proto;
  uint8_t   type;
  uint64_t  stamp;
  bool      ret   = 0;
  bool      rekey = 0;
  char      ips_a[IPV4_BUF];
  char      ips_b[IPV4_BUF];
  
  if(!BLK_VALIDSIZE(b_size)){
  #ifdef LOG_NETSPAM
    log_err("invalid packet length received (%zu bytes)",b_size);
  #endif
    goto stat;
  }

  r_size = RED_SIZEB(b_size);
  if(!find_key(b,r_size,&p,&pk)){
  #ifdef LOG_NETSPAM
    log_err("key not found, cannot decrypt packet");
  #endif
    goto stat;
  }

  if(p->flags & PEER_MUTE){
    goto stat;
  }

  r = b+BLK_RED_OFF;
  blk_deciph(r,b,r_size,pk->ciph_serp);

  l       = r+RED_LOAD_OFF;
  l_size  = RED_LOAD_SIZE(r_size);
  rxseq   = get_u32le (r+RED_SEQNO_OFF);
  proto   = get_u8le  (r+RED_PROTO_OFF);
  type    = get_u8le  (r+RED_TYPE_OFF);

  if(proto != PAK_PROTO){
    log_err("unsupported protocol version %u from peer %s",
            proto,p->id);
    goto clear;
  }

  if(pkey_self(pk,rxseq)){
  #ifdef LOG_NETSPAM
    log_err("reflected packet (seqno %08X)",rxseq);
  #endif
    goto clear;
  }

  if(!pkey_recv(pk,rxseq)){
  #ifdef LOG_NETSPAM
    log_err("duplicate/old packet (seqno %08X) from peer %s",
            rxseq,p->id);
  #endif
    goto stat;
  }

  stamp     = time_sec();
  pk->stamp = stamp;
  if(!matryoshka){
    addr        = addr_from_in(addr_in);
    addr.stamp  = stamp;
    peer_addr_auto(p,addr);
  }
  peer_edit(p);

  switch(type){
  case PAK_TYPE_IP:
    ret = ip_recv(p,l,l_size);
    break;

  case PAK_TYPE_FLOOD:
    if(matryoshka){
      log_err("flood packet inside flood packet received from peer %s",
              p->id);
      goto stat;
    }
    ret = flood_recv(p,l,l_size);
    break;

  case PAK_TYPE_PROD:
    ret = prod_recv(p,l,l_size,rxseq);
    break;

  case PAK_TYPE_ADDR:
    ret = addr_recv(p,l,l_size);
    break;

  case PAK_TYPE_KEYOFFER1:
  case PAK_TYPE_KEYOFFER2:
  case PAK_TYPE_KEYSLICE1:
  case PAK_TYPE_KEYSLICE2:
  case PAK_TYPE_KEYDONE1:
  case PAK_TYPE_KEYDONE2:
    ret   = rekey_recv(p,pk,type,l,l_size);
    rekey = 1;
    break;

  default:
    log_err("unknown greasypak type %u from peer %s",
            type,p->id);
    break;
  }

  if(ret && !rekey){
    rekey_recv(p,pk,type,l,l_size);
  }

clear:
#ifdef MEMSET_RED
  memset_ffs(r,0x00,RED_SIZE);
#endif
stat:
  if(ret){
    if(!matryoshka){
      state_grx(b_size);
      peer_grx(p,b_size);
    }
  }else{
    state_rrx(b_size);
    if(p) peer_rrx(p,b_size);
  }
  return ret;
}

