#include "daemon.h"
#include "internal.h"
#include "pak.h"
#include <config.h> //MEMSET_RED
#include "state/state.h"
#include "peer/peer.h"
#include "peer/key.h"
#include <io/io.h>
#include <net/ipv4.h>
#include <log/log.h>
#include <util/time.h>

static void discard(size_t b_size,peer_t *p){
  state_dtx(b_size);
  peer_dtx(p,b_size);
}

//peer_cansend() MUST OF been checked b4 calling this
static bool tx_blk(uint8_t    *b,
                   size_t      b_size,
                   peer_t     *p,
                   pkey_t     *pk,
                   addr_in_t   addr){

  size_t block_n = 0;

  assert(peer_cansend(p,0,0,0));

  while(1){
    switch(udp_sendto(udp_fd,addr,b,b_size,0)){
    case IO_OK:
      goto done;

    case IO_BLOCK:
      if(++block_n > BLOCK_MAX){
        log_err("udp socket blocked %zu times, discarding packet",
                block_n);
        goto fail;
      }
      sleep_nsec(BLOCK_WAIT);
      break;

    case IO_DOWN:
      goto fail;

    case IO_FAIL:
    default:
      log_err("failed to write to udp socket, discarding packet");
      goto fail;
    }
  }

done:
  pkey_used(p,pk);
  state_gtx(b_size);
  peer_gtx(p,b_size);
  return 1;

fail:
  discard(b_size,p);
  return 0;
}

//peer_cansend() MUST OF been checked b4 calling this
static bool tx_blk_load(uint8_t    *b,
                        size_t      load_size,
                        uint8_t     type,
                        peer_t     *p,
                        pkey_t     *pk,
                        addr_in_t   addr){

  size_t b_size;

  if(!blk_prepare(b,load_size,type,pk,&b_size)){
    discard(b_size,p);
    return 0;
  }

  if(!tx_blk(b,b_size,p,pk,addr)){
    return 0;
  }

  return 1;
}

//peer_cansend() MUST OF been checked b4 calling this
bool tx_load(uint8_t const *load,
             size_t         load_size,
             uint8_t        type,
             peer_t        *p,
             pkey_t        *pk,
             addr_in_t      addr){

  uint8_t   b[BLK_MAXSIZE];
  uint8_t  *r = b+BLK_RED_OFF;
  uint8_t  *l = r+RED_LOAD_OFF;

  if(load){
    load_copy(l,load,load_size);
  }else{
    load_size = 0;
  }

  return tx_blk_load(b,load_size,type,p,pk,addr);
}

bool tx_ipv4(uint8_t *ipak,size_t ipak_size){
  uint32_t    src;
  uint32_t    dst;
  char        ips_a       [IPV4_BUF];
  char        ips_b       [IPV4_BUF];
  uint8_t     b           [BLK_MAXSIZE];
  uint8_t    *r;
  uint8_t    *l;
  uint8_t    *pak;
  size_t      frag_size;
  size_t      frag_off;
  bool        more;
  peer_t     *p = 0;
  pkey_t     *pk;
  addr_t     *addr;
  char const *err;
  bool        ret = 0;

  if(!ver_check (ipak,          DIR_OUT)) goto clear;
  if(!len_check (ipak,ipak_size,DIR_OUT)) goto clear;
  if(!sum_check (ipak,          DIR_OUT)) goto clear;
  if(!ttl_dec   (ipak,          DIR_OUT)) goto clear;

  src = ipv4_src_get(ipak);
  dst = ipv4_dst_get(ipak);

  if(src != state.ip){
    log_err("outgoing source ip mismatch (%s != %s)",
            IP_A(src),IP_B(state.ip));
    goto clear;
  }

  if(dst == state.ip){
    ipv4_update_sum(ipak);
    tun_write(ipak,ipak_size);
    ret = 1;
    goto clear;
  }

  p = peer_lookup_ip(dst);
  if(!p){
    log_err("outgoing destination ip unknown (%s)",
            IP_A(dst));
    goto clear;
  }

  if(p->flags & PEER_MUTE){
    goto clear;
  }

  r         = b+BLK_RED_OFF;
  l         = r+RED_LOAD_OFF;
  pak       = l+IP_PAK_OFF;
  frag_off  = 0;
  more      = 1;

  do{
    switch(ipv4_frag(pak,
                     ipak,
                     IP_PAK_MAXSIZE,
                     ipak_size,
                     &frag_off,
                     &frag_size)){

    case IPV4_FRAG_LAST:
      more = 0; 
    case IPV4_FRAG_MORE:
      //must be cheked here cuz tx num can be exhausted inside this loop
      if(!peer_cansend(p,&pk,&addr,&err)){
        log_err("%s for peer %s, cannot send ipv4 packet",
                err,p->id);
        goto clear;
      }

      if(!tx_blk_load(b,
                      frag_size,
                      PAK_TYPE_IP,
                      p,
                      pk,
                      addr_to_in(*addr))){
        goto done;
      }

      state_ntx(frag_size);
      peer_ntx(p,frag_size);
      break;

    default:
      assert(0);
    }
  }while(more);

done:
  ret = 1;
clear:
#ifdef MEMSET_RED
  memset_ffs(ipak,0x00,ipak_size);
#endif
  if(!ret){
    state_rtx(ipak_size);
    if(p) peer_rtx(p,ipak_size);
  }
  return ret;
}

