#include "type.h"
#include "internal.h"
#include <util/util.h>
#include <util/hex.h>
#include <util/serial.h>

#include <string.h>

static bool hex_set(val_t v,uint8_t *out,size_t size,char const **err){
  if(!_stringp(v)){
    if(err) *err = "must be hex string";
    return 0;
  }

  if(strlen(str_data(v)) != HEX_LEN(size)){
    if(err) *err = "hex string must be correct length";
    return 0;
  }

  if(!hex_dec(out,size,str_data(v),0)){
    if(err) *err = "invalid hex string";
    memset_ffs(out,0x00,size);
    return 0;
  }

  return 1;
}

bool key_set(val_t v,uint8_t *key,char const **err){
  return hex_set(v,key,KEY_SIZE,err);  
}

bool crc_set(val_t v,uint32_t *crc,char const **err){
  uint8_t buf[CRC32_SIZE];

  if(!hex_set(v,buf,sizeof(buf),err)){
    return 0;
  }

  assert(sizeof(*crc) == sizeof(buf));
  memcpy(crc,buf,sizeof(buf));
  *crc = host32be(*crc);
  return 1;
}

static val_t keyhex_get(uint8_t const *k,bool crc){
  char hex[HEX_BUF(KEY_SIZE)];
  assert(KEY_SIZE >= CRC32_SIZE);

  if(crc){
    key_crc_hex(hex,sizeof(hex),k);
  }else{
    hex_enc(hex,sizeof(hex),k,KEY_SIZE);
  }

  return str(hex);
}

val_t kinfo_get(pkey_t const *k,char const *id,bool crc){
  val_t lst;

  lst = list(cons(sym("KEY" ),   keyhex_get(k->key.full,crc)),
             cons(sym("RXSEQ"),  itg(k->rxseq)),
             cons(sym("TXSEQ"),  itg(k->txseq)),
             cons(sym("STAMP"),  itg(k->stamp)));
  
  if(id) push(cons(sym("ID"),str(id)),lst);

  return cons(PP_ALIST,lst);
}

static bool key_aset(val_t        sx,
                     char const  *opt,
                     pkey_t      *out,
                     char const **err){

  val_t v;

  if(!cdrassoc(sx,opt,&v)){
    if(err) *err = "key missing";
    return 0;
  }

  if(!key_set(v,out->key.full,err)){
    return 0;
  }

  pkey_prep(out);
  return 1;
}

static bool seq_aset(val_t        sx,
                     char const  *opt,
                     uint32_t    *out,
                     char const **err){

  val_t   v;
  int64_t i;

  if(!cdrassoc(sx,opt,&v)){
    if(err) *err = "sequence number missing";
    return 0;
  }

  if(!_integerp(v)){
    if(err) *err = "sequence number must be integer";
    return 0;
  }
          
  i = itg_data(v);
  if(i < 0 || i > 0xFFFFFFFF){
    if(err) *err = "sequence number out of range";
    return 0;
  }

  *out = i;
  return 1;
}

static bool stamp_aset(val_t        sx,
                       char const  *opt,
                       uint64_t    *out,
                       char const **err){

  val_t   v;
  int64_t i;

  if(!cdrassoc(sx,opt,&v)){
    if(err) *err = "timestamp missing";
    return 0;
  }

  if(!_integerp(v)){
    if(err) *err = "timestamp number must be integer";
    return 0;
  }
          
  i = itg_data(v);
  if(i < 0){
    if(err) *err = "timestamp out of range";
    return 0;
  }

  *out = (uint64_t)i;
  return 1;
}

bool kinfo_set(val_t v,pkey_t *k,char const **err){
  if(!_listp(v)){
    if(err) *err = "key info must be alist";
    return 0;
  }

  if(!key_aset  (v,"KEY",   k,        err)) return 0;
  if(!seq_aset  (v,"RXSEQ", &k->rxseq,err)) return 0;
  if(!seq_aset  (v,"TXSEQ", &k->txseq,err)) return 0;
  if(!stamp_aset(v,"STAMP", &k->stamp,err)) return 0;

  return 1;
}

bool keys_set(val_t        v,
              pkey_t      *k,
              size_t      *k_n,
              size_t       k_size,
              char const **err){

  size_t n = 0;

  if(!_listp(v)){
    if(err) *err = "must be list";
    return 0;
  }

  while(!_null(v)){
    pkey_t *pk;

    if(n >= k_size){
      if(err) *err = "too many keys";
      goto clear;
    }

    pk = &k[n++];
    pkey_init(pk);
    if(!kinfo_set(car(v),pk,err)){
      goto clear;
    }

    v = cdr(v);
  }

  *k_n = n;
  return 1;

clear:
  memset_ffs(k,0x00,sizeof(*k)*n);
  return 0;
}

val_t keys_get(pkey_t const *k,size_t k_n,char const *id,bool crc){
  val_t v = NIL;
  for(size_t x = 0;x < k_n;x++){
    push(kinfo_get(k+x,id,crc),v);
  }
  return nreverse(v);
}

bool keys_aset(val_t        sx,
               char const  *opt,
               pkey_t      *out,
               size_t      *out_n,
               size_t       out_size,
               char const **err){

  val_t v;

  if(!cdrassoc(sx,opt,&v)){
    *out_n = 0;
    return 1;
  }

  return keys_set(v,out,out_n,out_size,err);
}

