#include "daemon.h"
#include "internal.h"
#include "pak.h"
#include "state/state.h"
#include "peer/peer.h"
#include <io/io.h>
#include <net/ipv4.h>
#include <util/time.h>
#include <log/log.h>

static char const *dir_str[] = {
  [DIR_IN]  = "incoming",
  [DIR_OUT] = "outgoing",
};

bool ver_check(uint8_t const *ipak,int dir){
  uint8_t ver = ipv4_ver_get(ipak);
  if(ver != IPV4_VERSION){
    log_err("%s ip packet unsupported version (%u)",
            dir_str[dir],ver);
    return 0;
  }
  return 1;
}

bool len_check(uint8_t const *ipak,size_t ipak_size,int dir){
  uint16_t len = ipv4_len_get(ipak);
  if(len != ipak_size){
    log_err("%s ipv4 length mismatch (%zu != %zu)",
            dir_str[dir],(size_t)len,ipak_size);
    return 0;
  }
  return 1;
}

bool sum_check(uint8_t const *ipak,int dir){
  uint16_t  sum   = ipv4_sum_get(ipak);
  uint16_t  suml  = ipv4_sum(ipak);
  if(sum != suml){
    log_err("%s ipv4 checksum mismatch (%04X != %04X)",
            dir_str[dir],sum,suml);
    return 0;
  }
  return 1;
}

bool ttl_dec(uint8_t *ipak,int dir){
  uint8_t ttl = ipv4_ttl_get(ipak);

  if(ttl == 0){
    log_err("%s ttl == 0",dir_str[dir]);
    return 0;
  }

  if(--ttl == 0){
    return 0;
  }

  ipv4_ttl_set(ipak,ttl);
  return 1;
}

void ipv4_update_sum(uint8_t *ipak){
  ipv4_sum_set(ipak,ipv4_sum(ipak));
}

bool key_gen(uint8_t *key){
  if(!rng_read(rng_slow_fd,key,KEY_SIZE)){
    log_err("failed to generate key, rng read failed");
    return 0;
  }

  if(entropy_missing(key,KEY_SIZE)){
    log_err("failed to generate key, entropy missing");
    return 0;
  }

  return 1;
}

void set_sockaddr_ip(void *sa,uint32_t ip){
  addr_in((struct sockaddr_in *)sa,ADDR_IN_CONST(ip,0));
}

static bool char_valid(char c){
  return (c >= 'a' && c <= 'z') ||
         (c >= 'A' && c <= 'Z') ||
         (c >= '0' && c <= '9') ||
         c == '_';
}

bool id_valid(char const *id,char const **err){
  size_t len = 0;

  if(!id[0]){
    if(err) *err = "cannot be empty";
    return 0;
  }

  for(char c;c = *id++;len++){
    if(char_valid(c)) continue;
    if(err) *err = "invalid character (use A-Za-z0-9_)";
    return 0;
  }

  if(len > ID_LEN){
    if(err) *err = "too long";
    return 0;
  }

  return 1;
}

bool tun_write(uint8_t const *ipak,size_t ipak_size){
  size_t block_n  = 0;
  size_t n        = 0;
  size_t w;

  while(n < ipak_size){
    switch(fd_write(dev_fd,ipak+n,ipak_size-n,&w)){
    case IO_OK:
      n += w;
      break;

    case IO_BLOCK:
      if(++block_n > BLOCK_MAX){
        log_err("tunnel device blocked %zu times, discarding packet",
                block_n);
        return 0;
      }
      sleep_nsec(BLOCK_WAIT);
      break;

    case IO_FAIL:
    default:
      log_err("failed to write to tunnel device, discarding packet");
      return 0;
    }
  }

  return 1;
}

void pkey_used(peer_t *p,pkey_t *pk){
  pkey_sent(pk);
  peer_edit(p);
  //TODO: omly do this if not alredy done x min ago
  if(pkey_txhalf(pk) && !rekey_active(&p->rekey)){
    rekey_initiate(p);
  }
}

void addr_init(addr_t *a){
  a->ip     = 0;
  a->port   = 0;
  a->stamp  = 0;
}

bool addr_null(addr_t a){
  return a.ip   == 0 ||
         a.port == 0;
}

bool addr_equal(addr_t a,addr_t b){
  return a.ip   == b.ip &&
         a.port == b.port;
}

addr_in_t addr_to_in(addr_t a){
  addr_in_t ai;
  ai.ip   = a.ip;
  ai.port = a.port;
  return ai;
}

addr_t addr_from_in(addr_in_t ai){
  addr_t a;
  a.ip    = ai.ip;
  a.port  = ai.port;
  a.stamp = 0;
  return a;
}

