#include "fs.h"
#include "internal.h"
#include <def.h>
#include <config.h> //LOG_IOSPAM
#include <util/util.h>
#include <util/string.h>
#include <log/log.h>

#include <errno.h>
#include <unistd.h>
#include <sys/stat.h>

bool fs_exists(char const *path){
  return access(path,F_OK) == 0;
}

bool fs_rename(char const *path_old,char const *path_new){
  if(rename(path_old,path_new) == 0){
#ifdef LOG_IOSPAM
    log_out("renamed %s -> %s",path_old,path_new);
#endif
    return 1;
  }else{
    log_errno("rename");
    log_err("failed to rename %s -> %s",path_old,path_new);
    return 0;
  }
}

bool fs_mkdir(char const *path,mode_t mode){
  if(mkdir(path,mode) == 0){
#ifdef LOG_IOSPAM
    log_out("directory created %s",path);
#endif
  }else if(errno != EEXIST){
    log_errno("mkdir");
    log_err("failed to create directory %s",path);
    return 0;
  }
  return 1;
}

bool fs_mkdir_p(char const *path,mode_t mode){
  char  buf[PATH_BUF];
  char *p = buf;
  char  t;

  strzcpy(buf,path,sizeof(buf));

  if(*p == '/'){
    p = strnotchr0(++p,'/');
  }

  do{
    p   = strchr0(p,'/');
    t   = *p;
    *p  = 0;
    if(!fs_mkdir(buf,mode)) return 0;
    *p  = t;
    p   = strnotchr0(++p,'/');
  }while(t);

  return 1;
}

bool fs_rm_f(char const *path){
  if(unlink(path) != 0){
    if(errno == ENOENT) return 1;
    log_errno("unlink");
    log_err("failed to delete file %s",path);
    return 0;
  }
#ifdef LOG_IOSPAM
  log_out("file deleted %s",path);
#endif
  return 1;
}

bool fs_rmdir_f(char const *path){
  if(rmdir(path) != 0){
    if(errno == ENOENT) return 1;
    log_errno("rmdir");
    log_err("failed to delete directory %s",path);
    return 0;
  }
#ifdef LOG_IOSPAM
  log_out("directory deleted %s",path);
#endif
  return 1;
}

bool fs_zero(char const *path){
  uint8_t   buf[4096];
  size_t    size;
  FILE     *f;

  f = fp_open(path,"rb+");
  if(!f) goto fail;

  if(!fp_size(f,&size)) goto close;

  BUF_ZERO(buf);

  while(size > 0){
    size_t w_size = size;
    if(w_size > sizeof(buf)){
      w_size = sizeof(buf);
    }
    if(!fp_write(f,buf,w_size)){
      return 0;
    }
    size -= w_size;
  }

  if(!fp_seek(f,0,SEEK_SET))  return 0;
  if(!fp_close(f))            return 0;

#ifdef LOG_IOSPAM
  log_out("file zeroized %s",path);
#endif
  return 1;

close:
  fp_close(f);
fail:
  return 0;
}

bool fs_chown(char const *path,uid_t uid,gid_t gid){
  if(chown(path,uid,gid) != 0){
    log_errno("chown");
    log_err("failed to change owner of %s",path);
    return 0;
  }
#ifdef LOG_IOSPAM
  log_out("changed owner of %s (uid=%d, gid=%d)",
          path,uid,gid);
#endif
  return 1;
}

bool fs_chmod(char const *path,mode_t mode){
  if(chmod(path,mode) != 0){
    log_errno("chmod");
    log_err("failed to change mode of %s",path);
    return 0;
  }
#ifdef LOG_IOSPAM
  log_out("changed mode of %s (mode=%04o)",
          path,mode);
#endif
  return 1;
}

bool fs_chdir(char const *path){
  if(chdir(path) != 0){
    log_errno("chdir");
    log_err("failed to change directory to %s",path);
    return 0;
  }
#ifdef LOG_IOSPAM
  log_out("changed directory to %s",path);
#endif
  return 1;
}

void disk_read(char const *path){
#ifdef LOG_IOSPAM
  log_out("file read %s",path);
#endif
}

void disk_written(char const *path){
#ifdef LOG_IOSPAM
  log_out("file written %s",path);
#endif
}

