#include "lisp.h"
#include "internal.h"
#include <util/util.h>
#include <util/string.h>

#include <stdio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <assert.h>

extern val_t        itg         (int64_t x);
extern val_t        chr         (char x);

extern char const  *str_data    (val_t x);
extern int64_t      itg_data    (val_t x);
extern char         chr_data    (val_t x);

extern val_t        caar        (val_t x);
extern val_t        cdar        (val_t x);
extern val_t        cadr        (val_t x);
extern val_t        cddr        (val_t x);

extern bool         _null       (val_t x);
extern bool         _consp      (val_t x);
extern bool         _listp      (val_t x);
extern bool         _symbolp    (val_t x);
extern bool         _stringp    (val_t x);
extern bool         _integerp   (val_t x);
extern bool         _characterp (val_t x);

typedef val_t     (*key_f)      (val_t);
typedef bool      (*test_f)     (val_t,val_t);

lisp_ctx_t         *_ctx = 0;

lisp_ctx_t *lisp_ctx(lisp_ctx_t *ctx){
  lisp_ctx_t *tmp;
  tmp   = _ctx;
  _ctx  = ctx;
  return tmp;
}

void lisp_init(void){
  _ctx->err_set   = 0;
  _ctx->cons_n    = 0;
  _ctx->sym_n     = 0;
  _ctx->str_n     = 0;
  _ctx->sym_nil   = sym("NIL"   );
  _ctx->sym_t     = sym("T"     );
  _ctx->sym_list  = sym(":LIST" );
  _ctx->sym_alist = sym(":ALIST");
}

void lisp_load(lisp_frame_t const *frame){
  #define ZERO(A) \
    memset_ffs(_ctx->A+frame->A##_n,\
               0x00,\
               sizeof(*(_ctx->A))*(_ctx->A##_n-frame->A##_n))

  ZERO(cons);
  ZERO(sym);
  ZERO(str);

  _ctx->cons_n  = frame->cons_n;
  _ctx->sym_n   = frame->sym_n;
  _ctx->str_n   = frame->str_n;
}

void lisp_store(lisp_frame_t *frame){
  frame->cons_n = _ctx->cons_n;
  frame->sym_n  = _ctx->sym_n;
  frame->str_n  = _ctx->str_n;
}

bool _lisp_try(void){
  _ctx->err_set = 1;
  return 1;
}

void lisp_untry(void){
  _ctx->err_set = 0;
}

char const *lisp_err(void){
  return _ctx->err_msg;
}

val_t sym(char const *name){
  val_t v;
  if(_find_symbol(name,&v)) return v;
  if(_ctx->sym_n+1 > _ctx->sym_size){
    eggog("out of symbol memory");
  }
  v.type      = TYPE_SYM;
  v.data.sym  = &_ctx->sym[_ctx->sym_n];
  if(!strzcpy_fail(v.data.sym->name,name,sizeof(v.data.sym->name))){
    eggog("symbol name too long");
  }
  _ctx->sym_n++;
  return v;
}

val_t str_begin(size_t *state){
  val_t v;
  *state      = _ctx->str_n;
  v.type      = TYPE_STR;
  v.data.str  = _ctx->str+_ctx->str_n;
  return v;
}

void str_abort(size_t state){
  _ctx->str_n = state;
}

void str_char(char c){
  if(_ctx->str_n+1 > _ctx->str_size){
    eggog("out of string memory");
  }
  _ctx->str[_ctx->str_n++] = c;
}

val_t str(char const *str){
  val_t   v;
  size_t  n = strlen(str)+1;
  if(_ctx->str_n+n > _ctx->str_size){
    eggog("out of string memory");
  }
  v.type        = TYPE_STR;
  v.data.str    = _ctx->str+_ctx->str_n;
  _ctx->str_n  += n;
  memcpy(v.data.str,str,n);
  return v;
}

val_t cons(val_t car,val_t cdr){
  val_t v;
  if(_ctx->cons_n+1 > _ctx->cons_size){
    eggog("out of cons memory");
  }
  v.type      = TYPE_CONS;
  v.data.cons = &_ctx->cons[_ctx->cons_n++];
  setf_car(v,car);
  setf_cdr(v,cdr);
  return v;
}

val_t car(val_t x){
  if(_null(x)) return NIL;
  check_type(x,TYPE_CONS);
  return x.data.cons->car;
}

val_t cdr(val_t x){
  if(_null(x)) return NIL;
  check_type(x,TYPE_CONS);
  return x.data.cons->cdr;
}

val_t setf_car(val_t x,val_t y){
  check_type(x,TYPE_CONS);
  x.data.cons->car = y;
  return y;
}

val_t setf_cdr(val_t x,val_t y){
  check_type(x,TYPE_CONS);
  x.data.cons->cdr = y;
  return y;
}

val_t last(val_t x){
  while(!_null(cdr(x))){
    x = cdr(x);
  }
  return x;
}

val_t nreverse(val_t x){
  val_t y;
  val_t z = NIL;
  while(!_null(x)){
    y = cdr(x);
    setf_cdr(x,z);
    z = x;
    x = y;
  }
  return z;
}

static val_t _assoc(val_t x,val_t a,test_f test,key_f acc){
  while(!_null(a)){
    if(test(x,acc(a))) return car(a);
    a = cdr(a);
  }
  return NIL;
}

val_t assoc(val_t x,val_t a){
  return _assoc(x,a,_eql,caar);
}

val_t assoc_equal(val_t x,val_t a){
  return _assoc(x,a,_equal,caar);
}

val_t rassoc(val_t x,val_t a){
  return _assoc(x,a,_eql,cdar);
}

val_t rassoc_equal(val_t x,val_t a){
  return _assoc(x,a,_equal,cdar);
}

bool _eql(val_t x,val_t y){
  if(x.type != y.type) return 0;
  switch(x.type){
  case TYPE_CONS: return x.data.cons == y.data.cons;
  case TYPE_SYM:  return x.data.sym  == y.data.sym;
  case TYPE_ITG:  return x.data.itg  == y.data.itg;
  case TYPE_CHR:  return x.data.chr  == y.data.chr;
  case TYPE_STR:  return x.data.str  == y.data.str;
  default:        return 0;
  }
}

bool _equal(val_t x,val_t y){
  if(x.type != y.type) return 0;
  switch(x.type){
  case TYPE_CONS:   return _equal(car(x),car(y)) &&
                           _equal(cdr(x),cdr(y));
  case TYPE_SYM:    
  case TYPE_ITG:    
  case TYPE_CHR:    return _eql(x,y);
  case TYPE_STR:    return strcmp(x.data.str,y.data.str) == 0;
  default:          return 0;
  }
}

size_t _length(val_t x){
  switch(x.type){
  case TYPE_CONS:
  case TYPE_SYM:
    ;size_t n = 0;
    while(!_null(x)){
      x = cdr(x);
      n++;
    }
    return n;

  case TYPE_STR:
    return strlen(x.data.str);

  default:
    eggog("wrong type retard");
    return 0;
  }
}

val_t _nth(size_t n,val_t x){
  if(_null(x)) return NIL;
  check_type(x,TYPE_CONS);
  if(n == 0) return car(x);
  return _nth(n-1,cdr(x));
}

val_t _list(int va,...){
  va_list   a;
  val_t     v;
  val_t     l;

  va_start(a,va);
  v = cons(va_arg(a,val_t),NIL);
  l = v;

  while(1){
    val_t u = va_arg(a,val_t);
    if(none(u)) break;
    l = setf_cdr(l,cons(u,NIL));
  }

  va_end(a);
  return v;
}

val_t _nconc(int va,...){
  va_list   a;
  val_t     v;
  val_t     l;

  va_start(a,va);

  do{
    v = va_arg(a,val_t);
    if(none(v)){
      va_end(a);
      return NIL;
    }
    check_list(v);
  }while(_null(v));
  l = last(v);

  while(1){
    val_t u = va_arg(a,val_t);
    if(none(u)) break;
    check_list(u);
    if(!_null(u)) l = last(setf_cdr(l,u));
  }

  va_end(a);
  return v;
}

bool _find_symbol(char const *name,val_t *out){
  for(size_t x = 0;x < _ctx->sym_n;x++){
    sym_t *sym = _ctx->sym+x;
    if(strcmp(name,sym->name) == 0){
      out->type     = TYPE_SYM;
      out->data.sym = sym;
      return 1;
    }
  }
  return 0;
}

char const *_symbol_name(val_t x){
  if(!_symbolp(x)) return 0;
  return x.data.sym->name;
}

void eggog(char const *msg){
  if(_ctx->err_set){
    _ctx->err_set = 0;
    strzcpy(_ctx->err_msg,msg,sizeof(_ctx->err_msg));
    longjmp(_ctx->err_jb,1);
  }
  fputs(msg,stderr);
  fputc('\n',stderr);
  exit(1);
}

void check_type(val_t x,unsigned type){
  if(x.type != type){
    eggog("wrong type dumas");
  }
}

void check_list(val_t x){
  if(!_listp(x)){
    eggog("not list dumas");
  }
}

