#ifndef LISP_H
#define LISP_H

#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <setjmp.h>
#include <assert.h>

#define LISP_SYM_LEN   15
#define LISP_SYM_BUF    (LISP_SYM_LEN+1)
#define LISP_ERR_BUF  256

#define NIL             (_ctx->sym_nil)
#define T               (_ctx->sym_t)
#define PP_LIST         (_ctx->sym_list)
#define PP_ALIST        (_ctx->sym_alist)

#define _NONE           (val_t){ .type = _TYPE_NONE }

#define lisp_try        if(setjmp(_ctx->err_jb) == 0 && _lisp_try())
#define lisp_catch      else
#define setf(X,Y)       ((X) = (Y))
#define push(X,Y)       ((Y) = cons(X,Y))
#define list(...)       _list   (0,__VA_ARGS__,_NONE)
#define nconc(...)      _nconc  (0,__VA_ARGS__,_NONE)

#define LISP_STATIC_CTX(NAME,CONS,SYM,STR) \
  static cons_t NAME##_cons [CONS];\
  static sym_t  NAME##_sym  [SYM];\
  static char   NAME##_str  [STR];\
  static lisp_ctx_t NAME = {\
    .cons       = NAME##_cons,\
    .cons_size  = sizeof(NAME##_cons)/sizeof(cons_t),\
    .sym        = NAME##_sym,\
    .sym_size   = sizeof(NAME##_sym)/sizeof(sym_t),\
    .str        = NAME##_str,\
    .str_size   = sizeof(NAME##_str)/sizeof(char),\
  }

enum{
  TYPE_CONS,
  TYPE_SYM,
  TYPE_ITG,
  TYPE_FLT,
  TYPE_CHR,
  TYPE_STR,

  _TYPE_NONE,
};

typedef union{
  struct cons      *cons;
  struct sym       *sym;
  int64_t           itg;
  char              chr;
  char             *str;
}data_t;

typedef struct val{
  unsigned          type;
  data_t            data;
}val_t;

typedef struct cons{
  val_t             car;
  val_t             cdr;
}cons_t;

typedef struct sym{
  char              name[LISP_SYM_BUF];
}sym_t;

typedef struct lisp_frame{
  size_t            cons_n;
  size_t            sym_n;
  size_t            str_n;
}lisp_frame_t;

typedef struct lisp_ctx{
  //memory
  cons_t           *cons;
  size_t            cons_size;
  sym_t            *sym;
  size_t            sym_size;
  char             *str;
  size_t            str_size;

  //state
  jmp_buf           err_jb;  
  bool              err_set;
  char              err_msg[LISP_ERR_BUF];
  size_t            cons_n;
  size_t            sym_n;
  size_t            str_n;

  val_t             sym_nil;
  val_t             sym_t;
  val_t             sym_list;
  val_t             sym_alist;

  //reader state
  lisp_frame_t      tmp;

  //printer state
  size_t            col;
}lisp_ctx_t;

extern lisp_ctx_t  *_ctx;

lisp_ctx_t         *lisp_ctx      (lisp_ctx_t *ctx);
void                lisp_init     (void);
void                lisp_load     (lisp_frame_t const *frame);
void                lisp_store    (lisp_frame_t       *frame);
bool                _lisp_try     (void);
void                lisp_untry    (void);
char const         *lisp_err      (void);

bool                lisp_read     (FILE *f,val_t *val);
size_t              lisp_print    (FILE *f,val_t  val);
size_t              lisp_pprint   (FILE *f,val_t  val);

val_t               sym           (char const *name);
val_t               str           (char const *str);
inline val_t        itg           (int64_t x);
inline val_t        chr           (char x);

inline char const  *str_data      (val_t x);
inline int64_t      itg_data      (val_t x);
inline char         chr_data      (val_t x);

val_t               cons          (val_t car,val_t cdr);
val_t               car           (val_t x);
val_t               cdr           (val_t x);
inline val_t        caar          (val_t x);
inline val_t        cdar          (val_t x);
inline val_t        cadr          (val_t x);
inline val_t        cddr          (val_t x);
val_t               setf_car      (val_t x,val_t y);
val_t               setf_cdr      (val_t x,val_t y);
val_t               last          (val_t x);
val_t               nreverse      (val_t x);
val_t               assoc         (val_t x,val_t a);
val_t               assoc_equal   (val_t x,val_t a);
val_t               rassoc        (val_t x,val_t a);
val_t               rassoc_equal  (val_t x,val_t a);

bool                _eql          (val_t x,val_t y);
bool                _equal        (val_t x,val_t y);
inline bool         _null         (val_t x);
inline bool         _consp        (val_t x);
inline bool         _listp        (val_t x);
inline bool         _symbolp      (val_t x);
inline bool         _stringp      (val_t x);
inline bool         _integerp     (val_t x);
inline bool         _characterp   (val_t x);
size_t              _length       (val_t x);
val_t               _nth          (size_t n,val_t x);
val_t               _list         (int va,...);
val_t               _nconc        (int va,...);
bool                _find_symbol  (char const *name,val_t *out);
char const         *_symbol_name  (val_t x);

inline val_t itg(int64_t x){
  val_t v;
  v.type      = TYPE_ITG;
  v.data.itg  = x;
  return v;
}

inline val_t chr(char x){
  val_t v;
  v.type      = TYPE_CHR;
  v.data.chr  = x;
  return v;
}

inline char const *str_data(val_t x){
  assert(x.type == TYPE_STR);
  return x.data.str;
}

inline int64_t itg_data(val_t x){
  assert(x.type == TYPE_ITG);
  return x.data.itg;
}

inline char chr_data(val_t x){
  assert(x.type == TYPE_CHR);
  return x.data.chr;
}

inline val_t caar(val_t x){
  return car(car(x));
}

inline val_t cdar(val_t x){
  return cdr(car(x));
}

inline val_t cadr(val_t x){
  return car(cdr(x));
}

inline val_t cddr(val_t x){
  return cdr(cdr(x));
}

inline bool _null(val_t x){
  return _eql(x,NIL);
}

inline bool _consp(val_t x){
  return x.type == TYPE_CONS;
}

inline bool _listp(val_t x){
  return _consp(x) || _null(x);
}

inline bool _symbolp(val_t x){
  return x.type == TYPE_SYM;
}

inline bool _stringp(val_t x){
  return x.type == TYPE_STR;
}

inline bool _integerp(val_t x){
  return x.type == TYPE_ITG;
}

inline bool _characterp(val_t x){
  return x.type == TYPE_CHR;
}

#endif

