#include "lisp.h"
#include "internal.h"
#include <inttypes.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>

enum{
  CONS_NORMAL,
  CONS_LIST,
  CONS_ALIST,
};

static size_t print_expr  (val_t v,bool p,FILE *f);
static size_t print_sym   (sym_t const *x,FILE *f);

static size_t set_col(size_t n){
  _ctx->col = n;
  return n;
}

static size_t inc_col(size_t n){
  _ctx->col += n;
  return n;
}

static size_t io_error(void){
  eggog("io write error");
  return 0;
}

static size_t print_c(char c,FILE *f){
  if(!f) return 1;
  if(fputc(c,f) == EOF){
    return io_error();
  }
  if(c == '\n'){
    set_col(0);
    return 1;
  }else{
    return inc_col(1);
  }
}

static size_t print_nl(size_t ind,FILE *f){
  size_t n = print_c('\n',f);
  for(size_t x = 0;x < ind;x++){
    n += print_c(' ',f);  
  }
  return n;
}

static size_t print_s(char const *s,FILE *f){
  if(!f) return strlen(s);
  if(fputs(s,f) == EOF){
    return io_error();
  }
  return inc_col(strlen(s));
}

static size_t print_f(FILE *f,char const *fmt,...){
  int n;
  va_list a;
  va_start(a,fmt);
  if(!f){
    char buf[1];
    n = vsnprintf(buf,sizeof(buf),fmt,a);
    va_end(a);
    return n;
  }else{
    n = vfprintf(f,fmt,a);
    va_end(a);
    if(n < 0){
      return io_error();
    }
    return inc_col(n);
  }
}

static size_t align_col(size_t c,FILE *f){
  size_t n = 0;
  while(_ctx->col < c){
    n += print_c(' ',f);
  }
  return n;
}

static bool cons_list(val_t x){
  return _eql(car(x),PP_LIST);
}

static bool cons_alist(val_t x){
  return _eql(car(x),PP_ALIST);
}

static bool cons_anylist(val_t x){
  return cons_list  (x) ||
         cons_alist (x);
}

static size_t print_alist_cons(val_t x,size_t align,FILE *f){
  size_t n = 0;
  check_type(x,TYPE_CONS);
  n += print_c('(',f);
  n += print_expr(car(x),1,f);
  n += align_col(align,f);
  n += print_s(" . ",f);
  n += print_expr(cdr(x),1,f);
  n += print_c(')',f);
  return n;
}

static size_t print_cons_int(val_t    x,
                             bool     p,
                             int      type,
                             size_t   align,
                             size_t   ind,
                             FILE    *f){

  size_t  n     = 0;
  val_t   car_x = car(x);
  val_t   cdr_x = cdr(x);

  if(p && type == CONS_ALIST){
    n += print_alist_cons(car_x,align,f);
  }else{
    n += print_expr(car_x,p,f);
  }

  if(_null(cdr_x)) return n;

  if(cdr_x.type == TYPE_CONS){
    if(type == CONS_NORMAL){
      n += print_c(' ',f);
    }else{
      n += print_nl(ind,f);
    }
    n += print_cons_int(cdr_x,p,type,align,ind,f);
  }else{
    n += print_s(" . ",f);
    n += print_expr(cdr_x,p,f);
  }

  return n;
}

static size_t alist_maxcar(val_t x){
  size_t max = 0;
  while(!_null(x)){
    size_t s = print_expr(caar(x),0,0);
    if(s > max) max = s;
    x = cdr(x);
  }
  return max;
}

static size_t print_cons(val_t x,bool p,FILE *f){
  size_t  n     = 0;
  int     type  = CONS_NORMAL;
  size_t  align = 0;

  if(p){
    if(cons_anylist(x) && _null(cdr(x))){
      return print_sym(NIL.data.sym,f);
    }else if(cons_list(x)){
      type  = CONS_LIST;
      x     = cdr(x);
    }else if(cons_alist(x)){
      type  = CONS_ALIST;
      x     = cdr(x);
      align = 1+_ctx->col+1+alist_maxcar(x); //'('+col+'('+caar()
    }
  }

  n += print_c('(',f);
  n += print_cons_int(x,p,type,align,_ctx->col,f);
  n += print_c(')',f);
  return n;
}

static bool char_lower(char c){
  return c >= 'a' && c <= 'z';
}

static bool sym_mustesc(char const *x){
  for(char c;c = *x++;){
    if(char_symesc(c) || char_lower(c)) return 1;
  }
  return 0;
}

static size_t print_sym(sym_t const *x,FILE *f){
  size_t      n     = 0;
  char const *name  = x->name;
  bool        esc   = sym_mustesc(name);
  if(esc) n += print_c('|',f);
  for(char c;c = *name++;){
    if(c == '\\' || c == '|'){
      n += print_c('\\',f);
    }
    n += print_c(c,f);
  }
  if(esc) n += print_c('|',f);
  return n;
}

static size_t print_itg(int64_t x,FILE *f){
  return print_f(f,"%"PRId64,x);
}

static size_t print_chr(char x,FILE *f){
  return print_f(f,"#\\%c",x);
}

static size_t print_str(char const *x,FILE *f){
  size_t n = 0;
  n += print_c('"',f);
  for(char c;c = *x++;){
    if(c == '"' || c == '\\'){
      n += print_c('\\',f);
    }
    n += print_c(c,f);
  }
  n += print_c('"',f);
  return n;
}

static size_t print_expr(val_t v,bool p,FILE *f){
#define CASE(U,L) case TYPE_##U: return print_##L(v.data.L,f);
  switch(v.type){
  case TYPE_CONS: return print_cons(v,p,f);
  CASE(SYM,sym);
  CASE(ITG,itg);
  CASE(CHR,chr);
  CASE(STR,str);
  default: return 0;
  }
}

size_t lisp_print(FILE *f,val_t val){
  _ctx->col = 0;
  return print_expr(val,0,f);
}

size_t lisp_pprint(FILE *f,val_t val){
  _ctx->col = 0;
  return print_expr(val,1,f);
}

