#include "lisp.h"
#include "internal.h"
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include <limits.h>

#define EXPR_BUF    LISP_SYM_BUF

typedef bool  (*char_f)     (char);

extern bool     char_white  (char c);
extern bool     char_symesc (char c);

static int      read_expr   (FILE *f,int lev,val_t *out);

enum{
  READ_FAIL,
  READ_SUCCESS,
  READ_CLOSE,
  READ_DOT,
};

static bool char_bin(char c){
  return c >= '0' && c <= '1';
}

static bool char_oct(char c){
  return c >= '0' && c <= '7';
}

static bool char_dec(char c){
  return c >= '0' && c <= '9';
}

static bool char_hex(char c){
  return (c >= '0' && c <= '9') ||
         (c >= 'a' && c <= 'f') ||
         (c >= 'A' && c <= 'F');
}

static bool char_suffix(char c){
  c = tolower(c);
  return c == 's' || 
         c == 'f' || 
         c == 'd' || 
         c == 'l' || 
         c == 'e';
}

static bool char_sign(char c){
  return c == '+' ||
         c == '-';
}

static void read_eggog(char const *msg){
  lisp_load(&_ctx->tmp);
  eggog(msg);
}

static bool char_eof(FILE *f,int c){
  if(c == EOF){
    if(ferror(f)) read_eggog("io read error");
    return 1;
  }
  return 0;
}

static int read_nowhite(FILE *f){
  int c;
  do{
    c = fgetc(f);
    if(char_eof(f,c)) return EOF;
  }while(char_white(c));
  return c;
}

static char fgetc_noeof(FILE *f,char const *egg){
  char c = fgetc(f);
  if(char_eof(f,c)) read_eggog(egg);
  return c;
}

static char *put_chr(char *dst,char src,char *dst_end,char const *egg){
  if(dst < dst_end-1){
    *dst++  = src;
    *dst    = 0;
    return dst;
  }else{
    read_eggog(egg);
    return 0;
  }
}

static int read_something(char *buf,size_t buf_size,FILE *f,char const *egg){
  char *b = buf;
  char *e = buf+buf_size;
  int   c;

  assert(b < e);
  *b = 0;

  while(1){
    c = fgetc(f);
    if(c == '\\'){
      c = fgetc_noeof(f,"\\ before eof");
    }else if(char_eof(f,c) || char_symesc(c)){
      if(c != EOF) ungetc(c,f);
      return c;
    }
    b = put_chr(b,c,e,egg);
  }
}

static val_t read_chr(FILE *f){
  char buf[sizeof("c")];
  read_something(buf,sizeof(buf),f,"char literal too long");
  return chr(*buf);
}

//TODO: riperoni these GREAT functions CACTUSFUKC LIBC IN ASHOLE
static int64_t read_strtoll(char const *buf,char **end,int base){
  int64_t r = strtoll(buf,end,base);
  if(r == LLONG_MIN ||
     r == LLONG_MAX){
    read_eggog("integer out of range");    
  }
  return r;
}

static val_t _read_itg(FILE *f,int base,char_f cf,char const *egg){
  char  buf[sizeof("18446744073709551616")]; 
  char *b = buf;
  char  c;
  read_something(buf,sizeof(buf),f,"integer literal too long");
  if(*b == '-') b++;
  while(c = *b++){
    if(!cf(c)) read_eggog(egg);
  }

  return itg(read_strtoll(buf,0,base));
}

static val_t read_bin(FILE *f){
  return _read_itg(f,2,char_bin,"invalid binary literal");
}

static val_t read_oct(FILE *f){
  return _read_itg(f,8,char_oct,"invalid octal literal");
}

static val_t read_hex(FILE *f){
  return _read_itg(f,16,char_hex,"invalid hexadecimal literal");
}

static int read_square(FILE *f,val_t *out){
  switch(tolower(fgetc_noeof(f,"unexpected eof"))){
  case '\\':  *out = read_chr(f); break;
  case 'b':   *out = read_bin(f); break;
  case 'o':   *out = read_oct(f); break;
  case 'x':   *out = read_hex(f); break;
  default:
    read_eggog("macro character not supported");
    return READ_FAIL;
  }
  return READ_SUCCESS;
}

static int read_str(FILE *f,val_t *out){
  size_t  state;
  val_t   v;
  int     c;

  v = str_begin(&state);

  while(1){
    c = fgetc(f);
    if(char_eof(f,c)) goto abort;
    if(c == '\\'){
      c = fgetc(f);
      if(char_eof(f,c)) goto abort;
    }else if(c == '"'){
      str_char(0);
      *out = v;
      return READ_SUCCESS;
    }
    str_char(c);
  }

abort:
  str_abort(state);
  return READ_FAIL;
}

static int read_escsym(FILE *f,val_t *out){
  char  buf[LISP_SYM_BUF];
  char *b = buf;
  char *e = buf+sizeof(buf);
  int   c;

  while(1){
    c = fgetc_noeof(f,"unterminated symbol");
    if(c == '\\'){
      c = fgetc_noeof(f,"unterminated symbol");
    }else if(c == '|'){
      *out = sym(buf);
      return READ_SUCCESS;
    }
    b = put_chr(b,c,e,"symbol too long");
  }
}

static val_t str_itg(char const *str){
  val_t v;
  v.type      = TYPE_ITG;
  v.data.itg  = read_strtoll(str,0,10);
  return v;
}

static int str_other(char *str,val_t *out){
  char       *s     = str;
  size_t      sign  = 0;
  size_t      len   = 0;
  size_t      dec   = 0;
  size_t      dot   = 0;
  size_t      suf   = 0;
  size_t      sufo  = -1;
  size_t      chr   = 0;
  size_t      num   = 0;
  char        c;
  char        last;

  while(c = *s++){
    if      (len == sufo+1 && char_sign(c)) sign++;
    else if (c == '.')                      dot++;
    else if (char_dec(c))                   dec++;
    else if (dot > 0 && char_suffix(c))     { suf++; sufo = len; }
    else                                    chr++;
    len++;
  }

  if(len == 0){
    read_eggog("no expression");
    return READ_FAIL;
  }

  num   = sign+dec;
  last  = str[len-1];

  if(dot == 1 && len == 1){
    return READ_DOT;
  }else if((dot == 0 && num < len) || dot > 1 || chr > 0 || sign == len){
    *out = sym(str);
  }else if(num == len || (num == len-1 && last == '.')){
    *out = str_itg(str);
  }else if(suf == 0 || (suf == 1 && char_dec(last))){
    read_eggog("floats not supported");
    return READ_FAIL;
  }else{
    read_eggog("invalid expression");
    return READ_FAIL;
  }

  return READ_SUCCESS;
}

static int read_other(FILE *f,val_t *out){
  char buf[EXPR_BUF];
  read_something(buf,sizeof(buf),f,"expression too long");
  for(char *b = buf;*b;b++){
    *b = toupper(*b);
  }
  return str_other(buf,out);
}

static int read_list(FILE *f,int lev,val_t *out){
  val_t v = NIL;
  int   d = -1;

  for(int x = 0;1;x++){
    val_t u;
    int   r;

    r = read_expr(f,lev,&u);
    if(r == READ_FAIL) return r;

    if(r == READ_CLOSE){
      if(d >= 0){
        if(d == x-1) read_eggog("nothing after . in list");
        if(d != x-2) read_eggog("more than one object follows . in list");
      }
      *out = v;
      return READ_SUCCESS;
    }

    if(r == READ_DOT){
      if(d >= 0) read_eggog("more than 1 . in list");
      if(x == 0) read_eggog("nothing before . in list");
      d = x; 
      continue;
    }

    if(d >= 0){
      setf_cdr(last(v),u);
    }else{
      v = nconc(v,list(u));
    }
  }
}

static int next_line(FILE *f){
  int c;
  do{
    c = fgetc(f);
  }while(!char_eof(f,c) && c != '\n');
  return c;
}

static int read_expr(FILE *f,int lev,val_t *out){
  int c = read_nowhite(f);

  switch(c){
  case EOF:
    return READ_FAIL;

  case ';':
    if(next_line(f) == EOF) return READ_FAIL;
    return read_expr(f,lev,out); 

  case '(':
    return read_list(f,lev+1,out);

  case ')':
    if(lev == 0) read_eggog("unmatched close parenthesis");
    return READ_CLOSE;

  case '#':
    return read_square(f,out);

  case '"':
    return read_str(f,out);

  case '|':
    return read_escsym(f,out);

  default:
    ungetc(c,f);
    return read_other(f,out);
  }
}

bool lisp_read(FILE *f,val_t *val){
  lisp_store(&_ctx->tmp);
  if(read_expr(f,0,val) == READ_SUCCESS){
    return 1; 
  }else{
    lisp_load(&_ctx->tmp);
    return 0; 
  }
}

