#define _POSIX_C_SOURCE 200809L

#include "log.h"
#include <def.h>
#include <util/time.h>
#include <util/util.h>
#include <lisp/lisp.h>

#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <sys/stat.h>

#define LOG_UFLAGS     0x0000FFFF
#define LOG_IFLAGS     (~LOG_UFLAGS)
#define LOG_FILE       0x00010000

#define STAMP_TIME_FMT          "%T"
#define STAMP_TIMEDATE_FMT      "%F %T"
#define STAMP_BUF             256

#define LOG_FP_MAX              8

typedef struct log_fp{
  FILE         *fp;
  uint32_t      flags;
}log_fp_t;

static log_fp_t log_fp[LOG_FP_MAX];
static size_t   log_fp_n = 0;

static void _log_add(FILE *fp,uint32_t flags){
  log_fp_t *l;
  if(log_fp_n >= LOG_FP_MAX) exit(1);
  l         = &log_fp[log_fp_n++];
  l->fp     = fp;
  l->flags  = flags;
}

void log_add(FILE *fp,uint32_t flags){
  _log_add(fp,flags & LOG_UFLAGS);
}

bool log_add_file(char const  *path,
                  uint32_t     flags,
                  mode_t       mode,
                  uid_t        uid,
                  gid_t        gid){

  FILE *fp;
  int   fd;

  fp = fopen(path,"ab");
  if(!fp){
    perror("fopen");
    fprintf(stderr,"failed to open log file %s\n",path);
    return 0;
  }

  fd = fileno(fp);
  if(fchown(fd,uid,gid) != 0){
    fclose(fp);
    perror("fchown");
    return 0;
  }

  if(fchmod(fd,mode) != 0){
    fclose(fp);
    perror("fchmod");
    return 0;
  }

  _log_add(fp,(flags & LOG_UFLAGS) | LOG_FILE);
  return 1;
}

void log_close(void){
  for(size_t x = 0;x < log_fp_n;x++){
    log_fp_t *l = log_fp+x;
    if(l->flags & LOG_FILE){
      fclose(l->fp);
    }
  }
  log_fp_n = 0;
}

static size_t time_str(char       *buf,
                       size_t      buf_size,
                       char const *fmt,
                       uint64_t    time){

  time_t      tt;
  struct tm  *info;
  size_t      n;

  tt    = time;
  info  = localtime(&tt);
  if(!info){
    buf[0] = 0;
    return 0;
  }

  n = strftime(buf,buf_size,fmt,info);
  if(!n){
    buf[0] = 0;
    return 0;
  }

  return n;
}

static void stamp_time(FILE *f,uint64_t time){
  char ts[STAMP_BUF];
  time_str(ts,sizeof(ts),STAMP_TIME_FMT" ",time);
  fputs(ts,f);
}

static void stamp_timedate(FILE *f,uint64_t time){
  char ts[STAMP_BUF];
  time_str(ts,sizeof(ts),STAMP_TIMEDATE_FMT" ",time);
  fputs(ts,f);
}

static void _log_v(uint32_t     flags,
                   log_fp_t    *l,
                   char const  *fmt,
                   va_list      args,
                   uint64_t     time){

  if(!(l->flags & flags)){
    return;
  }

  if(l->flags & LOG_TIMEDATE){
    stamp_timedate(l->fp,time);
  }else if(l->flags & LOG_TIME){
    stamp_time(l->fp,time);
  }

  if(flags & LOG_ERR){
    fputs("error: ",l->fp);
  }

  vfprintf(l->fp,fmt,args);
  fputc('\n',l->fp);
  fflush(l->fp);
}

void log_v(uint32_t flags,char const *fmt,va_list args){
  uint64_t time = time_sec();

  for(size_t x = 0;x < log_fp_n;x++){
    log_fp_t *l = log_fp+x;
    va_list args_cpy;
    va_copy(args_cpy,args);
    _log_v(flags & LOG_UFLAGS,l,fmt,args_cpy,time);
    va_end(args_cpy);
  }
}

void log_out(char const *fmt,...){
  va_list args;
  va_start(args,fmt);
  log_v(LOG_OUT,fmt,args);
  va_end(args);
}

void log_err(char const *fmt,...){
  va_list args;
  va_start(args,fmt);
  log_v(LOG_ERR,fmt,args);
  va_end(args);
}

void log_errno(char const *call){
  log_err("%s(): %s",call,strerror(errno));
}

void log_lisperr(void){
  log_err("lisp: %s",lisp_err());
}

