#ifndef IPV4_H
#define IPV4_H

#include <util/serial.h>
#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>
#include <assert.h>

#define IPV4_VERSION          4
#define IPV4_MTU          65535
#define IPV4_HDR_MINSIZE     20
#define IPV4_HDR_MAXSIZE     60

#define IPV4_VER_OFF          0
#define IPV4_IHL_OFF          0
#define IPV4_LEN_OFF          2
#define IPV4_FLA_OFF          6
#define IPV4_FRA_OFF          6
#define IPV4_TTL_OFF          8
#define IPV4_SUM_OFF         10
#define IPV4_SRC_OFF         12
#define IPV4_DST_OFF         16

#define IPV4_VER_SHIFT        4
#define IPV4_IHL_UNIT         4
#define IPV4_IHL_MASK      0x0F
#define IPV4_FLA_MASK      0xE0
#define IPV4_FLA_SHIFT        5
#define IPV4_FRA_MASK    0x1FFF
#define IPV4_FRA_UNIT         8

#define IPV4_FLA_MF        0x01
#define IPV4_FLA_DF        0x02

enum{
  IPV4_FRAG_DROP,
  IPV4_FRAG_MORE,
  IPV4_FRAG_LAST,
};

inline uint8_t  ipv4_ver_get  (uint8_t const *pak);
inline uint16_t ipv4_ihl_get  (uint8_t const *pak);
inline uint16_t ipv4_len_get  (uint8_t const *pak);
inline void     ipv4_len_set  (uint8_t       *pak,uint16_t len);
inline uint8_t  ipv4_fla_get  (uint8_t const *pak);
inline void     ipv4_fla_set  (uint8_t       *pak,uint8_t  fla);
inline uint16_t ipv4_fra_get  (uint8_t const *pak);
inline void     ipv4_fra_set  (uint8_t       *pak,uint16_t fra);
inline uint8_t  ipv4_ttl_get  (uint8_t const *pak);
inline void     ipv4_ttl_set  (uint8_t       *pak,uint8_t  sum);
inline uint16_t ipv4_sum_get  (uint8_t const *pak);
inline void     ipv4_sum_set  (uint8_t       *pak,uint16_t sum);
inline uint32_t ipv4_src_get  (uint8_t const *pak);
inline uint32_t ipv4_dst_get  (uint8_t const *pak);

uint16_t        ipv4_sum      (uint8_t const *pak);

int             ipv4_frag     (uint8_t       *dst,
                               uint8_t const *src,
                               size_t         dst_size,
                               size_t         src_size,
                               size_t        *_frag_off,
                               size_t        *dst_size_out);

inline uint8_t ipv4_ver_get(uint8_t const *pak){
  return get_u8be(pak+IPV4_VER_OFF)>>IPV4_VER_SHIFT;
}

inline uint16_t ipv4_ihl_get(uint8_t const *pak){
  uint8_t ihl = get_u8be(pak+IPV4_IHL_OFF);
  return (ihl & IPV4_IHL_MASK)*IPV4_IHL_UNIT;
}

inline uint16_t ipv4_len_get(uint8_t const *pak){
  return get_u16be(pak+IPV4_LEN_OFF);
}

inline void ipv4_len_set(uint8_t *pak,uint16_t len){
  set_u16be(pak+IPV4_LEN_OFF,len);
}

inline uint8_t ipv4_fla_get(uint8_t const *pak){
  uint8_t fla = get_u8be(pak+IPV4_FLA_OFF);
  return fla>>IPV4_FLA_SHIFT;
}

inline void ipv4_fla_set(uint8_t *pak,uint8_t fla){
  uint8_t old = get_u8be(pak+IPV4_FLA_OFF) & ~IPV4_FLA_MASK;
  set_u8be(pak+IPV4_FLA_OFF,(fla<<IPV4_FLA_SHIFT) | old);
}

inline uint16_t ipv4_fra_get(uint8_t const *pak){
  uint16_t fra = get_u16be(pak+IPV4_FRA_OFF);
  return (fra & IPV4_FRA_MASK)*IPV4_FRA_UNIT;
}

inline void ipv4_fra_set(uint8_t *pak,uint16_t fra){
  uint16_t old = get_u16be(pak+IPV4_FRA_OFF) & ~IPV4_FRA_MASK;
  assert(fra%IPV4_FRA_UNIT == 0);
  set_u16be(pak+IPV4_FRA_OFF,(fra/IPV4_FRA_UNIT) | old);
}

inline uint8_t ipv4_ttl_get(uint8_t const *pak){
  return get_u8be(pak+IPV4_TTL_OFF);
}

inline void ipv4_ttl_set(uint8_t *pak,uint8_t ttl){
  set_u8be(pak+IPV4_TTL_OFF,ttl);
}

inline uint16_t ipv4_sum_get(uint8_t const *pak){
  return get_u16be(pak+IPV4_SUM_OFF);
}

inline void ipv4_sum_set(uint8_t *pak,uint16_t sum){
  set_u16be(pak+IPV4_SUM_OFF,sum);
}

inline uint32_t ipv4_src_get(uint8_t const *pak){
  return get_u32be(pak+IPV4_SRC_OFF);
}

inline uint32_t ipv4_dst_get(uint8_t const *pak){
  return get_u32be(pak+IPV4_DST_OFF);
}

#endif

