#include "internal.h"
#include <io/io.h>
#include <util/util.h>
#include <log/log.h>

#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/un.h>
#include <netinet/in.h>

int sock_udp(void){
  int fd = socket(AF_INET,SOCK_DGRAM,0);
  if(fd == -1){
    log_errno("socket");
    log_err("failed to create udp socket");
  }
  return fd;
}

int sock_unix_dgram(void){
  int fd = socket(AF_UNIX,SOCK_DGRAM,0);
  if(fd == -1){
    log_errno("socket");
    log_err("failed to create unix socket");
  }
  return fd;
}

static bool sock_bind(int fd,void *_sa,size_t sa_size){
  struct sockaddr *sa = (struct sockaddr *)_sa;
  if(bind(fd,sa,sa_size) == -1){
    log_errno("bind");
    log_err("failed to bind socket");
    return 0;
  }
  return 1;
}

bool sock_bind_in(int fd,addr_in_t addr){
  struct sockaddr_in sa;
  size_t sa_size = addr_in(&sa,addr);
  return sock_bind(fd,&sa,sa_size);
}

bool sock_bind_un(int fd,char const *path){
  struct sockaddr_un sa;
  size_t sa_size = addr_un(&sa,path);
  return sock_bind(fd,&sa,sa_size);
}

static bool sock_nonblock(int fd){
  if(fcntl(fd,F_SETFL,O_NONBLOCK) == -1){
    log_errno("fcntl");
    log_err("failed to set socket options");
    return 0;
  }
  return 1;
}

static bool setopt(int fd,int level,int name,void *val,size_t len){
  if(setsockopt(fd,level,name,val,len) == -1){
    log_errno("setsockopt");
    log_err("failed to set socket options");
    return 0;
  }
  return 1;
}

static bool sock_reuseaddr(int fd,int enable){
  return setopt(fd,SOL_SOCKET,SO_REUSEADDR,&enable,sizeof(enable));
}

static bool sock_broadcast(int fd,int enable){
  return setopt(fd,SOL_SOCKET,SO_BROADCAST,&enable,sizeof(enable));
}

bool sock_flags(int fd,uint32_t flags){
  if((flags & NET_NONBLOCK) && !sock_nonblock(fd)){
    return 0;
  }

  if((flags & NET_REUSEADDR) && !sock_reuseaddr(fd,1)){
    return 0;
  }

  if((flags & NET_BROADCAST) && !sock_broadcast(fd,1)){
    return 0;
  }

  return 1;
}

bool sock_bufsize(int fd,int size){
  if(size == 0) return 1;
  if(!setopt(fd,SOL_SOCKET,SO_RCVBUF,&size,sizeof(size))) return 0;
  if(!setopt(fd,SOL_SOCKET,SO_SNDBUF,&size,sizeof(size))) return 0;
  return 1;
}

int sock_recvfrom(int            fd,
                  uint8_t       *data,
                  size_t         size,
                  void          *sa,
                  socklen_t     *sa_size,
                  size_t        *size_out){

  ssize_t n = recvfrom(fd,
                       data,
                       size,
                       0,
                       (struct sockaddr *)sa,
                       sa_size);

  if(n == -1){
    if(errno_block()) return IO_BLOCK;
    log_errno("recvfrom");
    log_err("failed to receive data from socket");
    return IO_FAIL;
  }
  if(size_out){
    *size_out = n;
  }
  return IO_OK;
}

int sock_sendto(int            fd,
                uint8_t const *data,
                size_t         size,
                void          *sa,
                socklen_t      sa_size,
                size_t        *size_out){

  ssize_t n = sendto(fd,  
                     data,
                     size,
                     0,
                     (struct sockaddr const *)sa,
                     sa_size);

  if(n == -1){
    if(errno_block()) return IO_BLOCK;
    if(errno_down())  return IO_DOWN;
    log_errno("sendto");
    log_err("failed to send data to socket");
    return IO_FAIL;
  }
  if(size_out){
    *size_out = n;
  }
  return IO_OK;
}

