#include "util.h"

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#define SUBNET_LOCAL_A SUBNET_CONST(IPV4_CONST( 10,  0,  0,  0), 8)
#define SUBNET_LOCAL_B SUBNET_CONST(IPV4_CONST(172, 16,  0,  0),12)
#define SUBNET_LOCAL_C SUBNET_CONST(IPV4_CONST(192,168,  0,  0),16)

bool addr_in_equal(addr_in_t a,addr_in_t b){
  return a.ip   == b.ip &&
         a.port == b.port;
}

void addr_in_fmt(char *buf,size_t buf_size,addr_in_t addr){
  uint8_t a = addr.ip >> 24;
  uint8_t b = addr.ip >> 16;
  uint8_t c = addr.ip >>  8;
  uint8_t d = addr.ip >>  0;
  snprintf(buf,buf_size,"%u.%u.%u.%u:%u",a,b,c,d,addr.port);
}

char *ipv4_fmt(char *buf,size_t buf_size,uint32_t ip){
  uint8_t a = ip >> 24;
  uint8_t b = ip >> 16;
  uint8_t c = ip >>  8;
  uint8_t d = ip >>  0;
  snprintf(buf,buf_size,"%u.%u.%u.%u",a,b,c,d);
  return buf;
}

static bool parse_byte(char const **b,uint32_t *ip,char t,size_t shl){
  char *e;
  long  i;

  i = strtol(*b,&e,10);
  if(*e != t)   return 0;
  if(i < 0x00)  return 0;
  if(i > 0xFF)  return 0;

  *ip  |= i << shl;
  *b    = e+1;
  return 1;
}

bool ipv4_parse(char const *buf,uint32_t *ip_out){
  char const *b   = buf;
  uint32_t    ip  = 0x00000000;
  if(!parse_byte(&b,&ip,'.',24)) return 0;
  if(!parse_byte(&b,&ip,'.',16)) return 0;
  if(!parse_byte(&b,&ip,'.', 8)) return 0;
  if(!parse_byte(&b,&ip,0,   0)) return 0;
  *ip_out = ip;
  return 1;
}

bool ipv4_mask(uint32_t ip,unsigned *mask_out){
  unsigned mask = 0;
  while(ip & 0x80000000){
    ip <<= 1;
    mask++;
  }
  if(ip != 0x00000000) return 0;
  *mask_out = mask;
  return 1;
}

uint32_t ipv4_maskip(unsigned mask){
  assert(mask <= 32);
  uint32_t ip = 0x00000000;
  while(mask--){
    ip  >>= 1;
    ip   |= 0x80000000;
  }
  return ip;
}

bool ipv4_insubnet(uint32_t ip,subnet_t p){
  uint32_t maskip = ipv4_maskip(p.mask);
  return (ip & maskip) == (p.ip & maskip);
}

bool ipv4_local(uint32_t ip){
  return ipv4_insubnet(ip,SUBNET_LOCAL_A) ||
         ipv4_insubnet(ip,SUBNET_LOCAL_B) ||
         ipv4_insubnet(ip,SUBNET_LOCAL_C);
}

char *subnet_fmt(char *buf,size_t buf_size,subnet_t p){
  uint8_t a = p.ip >> 24;
  uint8_t b = p.ip >> 16;
  uint8_t c = p.ip >>  8;
  uint8_t d = p.ip >>  0;
  snprintf(buf,buf_size,"%u.%u.%u.%u/%u",a,b,c,d,p.mask);
  return buf;
}

static bool parse_mask(char const **b,unsigned *mask,char t){
  char *e;
  long  i;

  i = strtol(*b,&e,10);
  if(*e != t) return 0;
  if(i <  0)  return 0;
  if(i > 32)  return 0;

  *mask = i;
  *b    = e+1;
  return 1;
}

bool subnet_parse(char const *buf,subnet_t *p_out){
  char const *b = buf;
  subnet_t    p = SUBNET_CONST(0x00000000,0);
  if(!parse_byte(&b,&p.ip,  '.',24)) return 0;
  if(!parse_byte(&b,&p.ip,  '.',16)) return 0;
  if(!parse_byte(&b,&p.ip,  '.', 8)) return 0;
  if(!parse_byte(&b,&p.ip,  '/', 0)) return 0;
  if(!parse_mask(&b,&p.mask,0     )) return 0;
  p.ip   &= ipv4_maskip(p.mask);
  *p_out  = p;
  return 1;
}

