#define __BSD_VISIBLE     1
#define _DEFAULT_SOURCE   1

#include "client.h"
#include "internal.h"
#include <def.h>
#include <io/io.h>
#include <io/fs.h>
#include <net/unix.h>
#include <net/util.h>
#include <util/util.h>
#include <util/string.h>
#include <lisp/lisp.h>
#include <log/log.h>

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <poll.h>
#include <fcntl.h>

#ifndef INFTIM
#define INFTIM -1
#endif

static bool     running;
static int      exit_code;
static int      cli_fd;
static uint8_t  sibuf     [RPC_BUF];
static size_t   sibuf_n;

enum{
  IDX_STDIN,
  POLL_STDIN_N,
};

enum{
  IDX_CLI,
  POLL_CLI_N,
};

static struct pollfd poll_stdin [POLL_STDIN_N];
static struct pollfd poll_cli   [POLL_CLI_N];

static void poll_init_in(struct pollfd *fds,size_t idx,int fd){
  fds[idx].fd     = fd;
  fds[idx].events = POLLIN;
}

static bool poll_readable(struct pollfd *fds,size_t idx){
  return fds[idx].revents & POLLIN;
}

static int cli_read(void){
  char    addr  [UNIX_PATH_BUF];
  uint8_t buf   [RPC_BUF];
  size_t  size;

  switch(unix_recvfrom(cli_fd,
                       addr,
                       sizeof(addr),
                       buf,
                       sizeof(buf),
                       &size)){

  case IO_OK:
    if(fd_write(STDOUT_FILENO,buf,size,0) != IO_OK){
      return IO_FAIL;
    }
    return IO_OK;

  case IO_BLOCK:
    return IO_BLOCK;

  case IO_FAIL:
  default:
    return IO_FAIL;
  }
}

static bool rpc_command(uint8_t const *cmd,size_t cmd_size){
  if(unix_sendto(cli_fd,
                 RPCS_PATH,
                 cmd,
                 cmd_size,
                 0) != IO_OK){
    return 0;
  }

  while(running){
    switch(poll(poll_cli,POLL_CLI_N,INFTIM)){
    case -1:
      if(errno == EINTR){
        running = 0;
        return 1;
      }
      log_errno("poll");
      log_err("failed to poll file descriptors");
      return 0;

    case 0:
      break;

    default:
      if(poll_readable(poll_cli,IDX_CLI)){
        switch(cli_read()){
        case IO_OK:
          return 1;

        case IO_FAIL:
          running = 0;
          return 0;

        case IO_BLOCK:
          break;
        }
      }
      break;
    }
  }

  return 1;
}

static FILE *sibuf_open(void){
  FILE *f = fmemopen(sibuf,sibuf_n,"rb");
  if(!f){
    log_errno("fmemopen");
    log_err("failed to open memory stream");
    return 0;
  }
  return f;
}

static void riperoni(size_t n){
  assert(n <= sibuf_n);
  sibuf_n -= n;
  memmove(sibuf,sibuf+n,sibuf_n);
}

static bool shit(void){
  lisp_frame_t  frame;
  FILE         *f = 0;
  bool          r = 0;

  if(sibuf_n >= RPC_BUF-1){
    log_err("s-expression too long");
    goto ret;
  }

  f = sibuf_open();
  if(!f) goto ret;

  lisp_store(&frame);

  lisp_try{
    val_t sx;
    while(lisp_read(f,&sx)){
      size_t n = ftell(f);
      fclose(f);
      f = 0;
        
      if(!rpc_command(sibuf,n)){
        goto pop;
      }

      riperoni(n);

      f = sibuf_open();
      if(!f) goto pop;
    }
  }lisp_catch{
    log_lisperr();
    goto close;
  }

  r = 1;
close:
  if(f) fclose(f);
pop:
  lisp_load(&frame);
ret:
  return r;
}

static bool stdin_read(void){
  size_t size;

  switch(fd_read(STDIN_FILENO,
                 sibuf+sibuf_n,
                 sizeof(sibuf)-sibuf_n,
                 &size)){

  case IO_OK:
    if(size == 0){
      exit_code = EXIT_SUCCESS;
      return 0;
    }
    sibuf_n += size;
    if(!shit()){
      exit_code = EXIT_FAILURE;
      return 0;
    }
    return 1;

  case IO_FAIL:
    exit_code = EXIT_FAILURE;
    return 0;

  case IO_BLOCK:
  default:
    return 1;
  }
}

int rpcc_run(void){
  char rpcc_dir   [PATH_BUF];
  char rpcc_path  [PATH_BUF];

  exit_code = EXIT_FAILURE;
  sibuf_n   = 0;

  strzcpy(rpcc_dir,RPCC_MKTEMP,sizeof(rpcc_dir));
  if(!mkdtemp(rpcc_dir)){
    log_errno("mkdtemp");
    log_err("failed to create temporary directory %s",rpcc_dir);
    goto ret;
  }
  
  snprintf(rpcc_path,sizeof(rpcc_path),"%s/%s",rpcc_dir,RPCC_PATH);

  cli_fd = unix_bind(rpcc_path,NET_NONBLOCK,RPC_BUF);
  if(cli_fd == -1){
    fs_rm_f(rpcc_path);
    goto rmdir;
  }

  poll_init_in(poll_stdin,IDX_STDIN, STDIN_FILENO);
  poll_init_in(poll_cli,  IDX_CLI,   cli_fd);

  running = 1;
  while(running){
    switch(poll(poll_stdin,POLL_STDIN_N,INFTIM)){
    case -1:
      if(errno == EINTR){
        running = 0;
        break;
      }
      log_errno("poll");
      log_err("failed to poll file descriptors");
      break;

    case 0:
      break;

    default:
      if(poll_readable(poll_stdin,IDX_STDIN) && !stdin_read()){
        running = 0;
      }
      break;
    }
  }

close_cli:
  close(cli_fd);
  fs_rm_f(rpcc_path);
rmdir:
  fs_rmdir_f(rpcc_dir);
ret:
  return exit_code;
}

void rpcc_stop(int code){
  exit_code = code;
  running   = 0;
}

