#define _POSIX_C_SOURCE 200809L

#include "server.h"
#include "internal.h"
#include <net/unix.h>
#include <net/util.h>
#include <io/io.h>
#include <io/fs.h>
#include <util/util.h>
#include <log/log.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <sys/stat.h>

int                 rpc_fd;

val_t               sym_err;
val_t               sym_ok;
val_t               val_ok;

static rpc_cmd_t   *rpc_cmd   = 0;
static size_t       rpc_cmd_n = 0;
static char const  *rpc_path  = 0;

static void         rpc_eggog(char const *addr,char const *msg);

static rpc_cmd_t *lookup_cmd(val_t cmd){
  for(size_t x = 0;x < rpc_cmd_n;x++){
    rpc_cmd_t *c = rpc_cmd+x;
    if(_eql(cmd,c->sym)) return c;
  }
  return 0;
}

bool rpcs_init(char const *path,rpc_cmd_t *cmd,size_t cmd_n){
  rpc_cmd   = cmd;
  rpc_cmd_n = cmd_n;
  rpc_path  = path;

  fs_rm_f(rpc_path);
  rpc_fd = unix_bind(rpc_path,NET_NONBLOCK,RPC_BUF);
  if(rpc_fd == -1) return 0;

  log_out("bound unix socket %s",rpc_path);

  lisp_try{
    sym_err = sym("ERR");
    sym_ok  = sym("OK");
    val_ok  = list(sym_ok);
  }lisp_catch{
    log_lisperr();
    return 0;
  }

  return 1;
}

bool rpcs_exit(void){
  if(close(rpc_fd) != 0){
    log_errno("close");
    log_err("failed to close unix socket");
    return 0;
  }
  log_out("closed unix socket");
  fs_rm_f(rpc_path);
  return 1;
}

static val_t rpc_exec(val_t val){
  val_t       cmd;
  val_t       arg;
  rpc_cmd_t  *c;

  if(!_consp(val)){
    return rpc_err("command must be cons");
  }
    
  cmd = car(val);
  arg = cdr(val);

  if(!_symbolp(cmd)){
    return rpc_err("car of command must be symbol");
  }

  c = lookup_cmd(cmd);
  if(c) return c->func(arg,_length(arg));

  return rpc_err("unknown rpc command %s",_symbol_name(cmd));
}

static bool rpc_read(char const  *addr,
                     uint8_t     *sexp,
                     size_t       sexp_size,
                     val_t       *val){

  val_t   tmp;
  int     r = 1;
  FILE   *f = fmemopen(sexp,sexp_size,"rb");
  if(!f){
    log_errno("fmemopen");
    log_err("failed to open memory stream");
    rpc_eggog(addr,"failed to open memory stream");
    return 0;
  }

  if(!lisp_read(f,val)){
    rpc_eggog(addr,"incomplete s-expression received");
    r = 0;
    goto close;
  }

  if(lisp_read(f,&tmp)){
    rpc_eggog(addr,"multiple s-expressions received");
    r = 0;
    goto close;
  }

close:
  fclose(f);
  return r;
}

static void rpc_print(char const *addr,
                      uint8_t    *buf,
                      size_t      buf_size,
                      val_t       val){

  size_t  size;
  FILE   *f = fmemopen(buf,buf_size,"wb");
  if(!f){
    log_errno("fmemopen");
    log_err("failed to open memory stream");
    rpc_eggog(addr,"failed to open memory stream");
    return;
  }

  size = lisp_pprint(f,val);

  if(fputc('\n',f) == EOF){
    rpc_eggog(addr,"rpc out buffer overflow");
    return;
  }
  size++;

  if(fclose(f) == EOF){
    rpc_eggog(addr,"failed to close rpc out buffer");
    return;
  }

  if(size >= RPC_BUF-1){
    rpc_eggog(addr,"response s-expression too long");
    return;
  }

  if(unix_sendto(rpc_fd,addr,buf,size,0) != IO_OK){
    log_err("failed to send to rpc socket");  
  }
}

static void rpc_eggog(char const *addr,char const *msg){
  lisp_frame_t frame;
  lisp_store(&frame);
  lisp_try{
    uint8_t   buf[ERR_BUF];
    size_t    size;
    val_t     val = rpc_err(msg);

    rpc_print(addr,buf,sizeof(buf),val);
  }lisp_catch{
    log_lisperr();
  }
  lisp_load(&frame);
}

static void rpc_respond(char const *addr,
                        uint8_t    *buf,
                        size_t      buf_size,
                        size_t      size){

  lisp_frame_t frame;
  lisp_store(&frame);

  if(size == 0){
    rpc_eggog(addr,"empty datagram recevied");
    goto pop;
  }

  if(size >= RPC_BUF-1){
    rpc_eggog(addr,"s-expression too long");
    goto pop;
  }

  lisp_try{
    val_t val_in;
    val_t val_out;
    if(!rpc_read(addr,buf,size,&val_in)) goto pop;
    val_out = rpc_exec(val_in);
    rpc_print(addr,buf,buf_size,val_out);
  }lisp_catch{
    log_lisperr();
    rpc_eggog(addr,lisp_err());
  }

pop:
  lisp_load(&frame);
}

void rpcs_recv(void){
  uint8_t buf[RPC_BUF];
  size_t  size;
  char    addr[UNIX_PATH_BUF];

  switch(unix_recvfrom(rpc_fd,
                       addr,
                       sizeof(addr),
                       buf,
                       sizeof(buf),
                       &size)){

  case IO_OK:
    rpc_respond(addr,buf,sizeof(buf),size);
    break;

  case IO_FAIL:
    log_err("failed to receive from rpc socket");
    break;

  case IO_BLOCK:
    break;
  }
}

val_t rpc_err(char const *fmt,...){
  char msg[1024];
  va_list va;
  va_start(va,fmt);
  vsnprintf(msg,sizeof(msg),fmt,va);
  va_end(va);
  return list(sym_err,str(msg));
}

